import types
from typing import Any, Dict, Optional

import pytest

from acapy_agent.database_manager.dbstore import DBStoreError, DBStoreErrorCode


class FakeDBStoreHandle:
    def __init__(self):
        self._rows: Dict[tuple[str, str], Dict[str, Any]] = {}

    def insert(self, category: str, name: str, value: str, tags: Optional[dict] = None):
        key = (category, name)
        if key in self._rows:
            raise DBStoreError(DBStoreErrorCode.DUPLICATE, "duplicate")
        self._rows[key] = {
            "category": category,
            "name": name,
            "value": value,
            "tags": tags or {},
        }

    def fetch(self, category: str, name: str, for_update: bool = False):
        row = self._rows.get((category, name))
        if not row:
            return None
        return types.SimpleNamespace(
            category=row["category"],
            name=row["name"],
            value=row["value"],
            tags=row["tags"],
        )

    def replace(self, category: str, name: str, value: str, tags: dict):
        if (category, name) not in self._rows:
            raise DBStoreError(DBStoreErrorCode.NOT_FOUND, "not found")
        self._rows[(category, name)] = {
            "category": category,
            "name": name,
            "value": value,
            "tags": tags,
        }

    def remove(self, category: str, name: str):
        if (category, name) not in self._rows:
            raise DBStoreError(DBStoreErrorCode.NOT_FOUND, "not found")
        del self._rows[(category, name)]

    async def fetch_all(self, category: str, tag_filter: Optional[dict] = None, **kwargs):
        # simple filter implementation
        for (cat, _), row in self._rows.items():
            if cat != category:
                continue
            tags = row["tags"] or {}
            ok = True
            for k, v in (tag_filter or {}).items():
                if tags.get(k) != v:
                    ok = False
                    break
            if ok:
                yield types.SimpleNamespace(
                    category=row["category"],
                    name=row["name"],
                    value=row["value"],
                    tags=row["tags"],
                )

    def remove_all(self, category: str, tag_filter: Optional[dict] = None):
        to_delete = []
        for (cat, name), row in self._rows.items():
            if cat != category:
                continue
            tags = row["tags"] or {}
            ok = True
            for k, v in (tag_filter or {}).items():
                if tags.get(k) != v:
                    ok = False
                    break
            if ok:
                to_delete.append((cat, name))
        for key in to_delete:
            del self._rows[key]


class FakeStoreSession:
    def __init__(self, handle):
        self.handle = handle

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return False

    async def fetch(self, *args, **kwargs):
        return await self.handle.fetch(*args, **kwargs)

    async def replace(self, *args, **kwargs):
        return await self.handle.replace(*args, **kwargs)

    async def insert(self, *args, **kwargs):
        return await self.handle.insert(*args, **kwargs)

    async def remove(self, *args, **kwargs):
        return await self.handle.remove(*args, **kwargs)

    async def fetch_all(self, *args, **kwargs):
        results = []
        async for row in self.handle.fetch_all(*args, **kwargs):
            results.append(row)
        return results

    async def remove_all(self, *args, **kwargs):
        return await self.handle.remove_all(*args, **kwargs)


class FakeProfile:
    def __init__(self):
        self._handle = FakeDBStoreHandle()
        self.store = self
        self.dbstore_handle = self._handle
        self.profile = types.SimpleNamespace(name="p")
        self.name = "p"

    def session(self):
        return FakeStoreSession(self._handle)

    @property
    def opened(self):
        return types.SimpleNamespace(db_store=self)

    def scan(
        self,
        *,
        category: str,
        tag_filter: Optional[dict] = None,
        limit: int = 10,
        offset: int = 0,
        profile: Optional[str] = None,
        order_by: Optional[str] = None,
        descending: bool = False,
    ):
        async def _gen():
            rows = []
            for (cat, name), row in self._handle._rows.items():
                if cat != category:
                    continue
                tags = row["tags"] or {}
                ok = True
                for k, v in (tag_filter or {}).items():
                    if tags.get(k) != v:
                        ok = False
                        break
                if ok:
                    rows.append(
                        types.SimpleNamespace(
                            category=cat, name=name, value=row["value"], tags=row["tags"]
                        )
                    )
            if order_by == "name":
                rows.sort(key=lambda r: r.name, reverse=descending)
            else:
                rows.sort(key=lambda r: r.name)
            _offset = offset or 0
            _limit = limit or len(rows)
            sliced = rows[_offset : _offset + _limit]
            for r in sliced:
                yield r

        return _gen()

    def scan_keyset(
        self,
        *,
        category: str,
        tag_filter: Optional[dict] = None,
        last_id: Optional[int] = None,
        limit: int = 10,
        profile: Optional[str] = None,
        order_by: Optional[str] = None,
        descending: bool = False,
    ):
        async def _gen():
            rows = [
                types.SimpleNamespace(
                    category=cat, name=name, value=row["value"], tags=row["tags"]
                )
                for (cat, name), row in self._handle._rows.items()
                if cat == category
            ]
            rows.sort(key=lambda r: r.name, reverse=descending)
            start = last_id or 0
            end = start + limit
            for r in rows[start:end]:
                yield r

        return _gen()


@pytest.mark.asyncio
async def test_add_get_update_delete_record_roundtrip():
    from acapy_agent.storage.kanon_storage import KanonStorage
    from acapy_agent.storage.record import StorageRecord

    profile = FakeProfile()
    storage = KanonStorage(profile)

    record = StorageRecord(type="config", id="pub", value='{"did": "D"}', tags={"k": "v"})

    await storage.add_record(record)

    got = await storage.get_record("config", "pub")
    assert got.value == record.value
    assert got.tags == {"k": "v"}

    await storage.update_record(record, value='{"did": "D2"}', tags={"k": "v2"})

    got2 = await storage.get_record("config", "pub")
    assert got2.value == '{"did": "D2"}'
    assert got2.tags == {"k": "v2"}

    # find_record path
    found = await storage.find_record("config", {"k": "v2"})
    assert found.id == "pub"

    # delete
    await storage.delete_record(record)
    with pytest.raises(Exception):
        await storage.get_record("config", "pub")


@pytest.mark.asyncio
async def test_find_paginated_records_and_keyset():
    from acapy_agent.storage.kanon_storage import KanonStorage
    from acapy_agent.storage.record import StorageRecord

    profile = FakeProfile()
    storage = KanonStorage(profile)

    for i in range(5):
        rec = StorageRecord(
            type="config",
            id=f"id{i}",
            value=f"v{i}",
            tags={"idx": str(i)},
        )
        await storage.add_record(rec)

    page = await storage.find_paginated_records(
        "config", limit=2, offset=1, order_by="name", descending=False
    )
    assert [r.id for r in page] == ["id1", "id2"]

    page1 = await storage.find_paginated_records_keyset(
        "config", limit=2, order_by="name", descending=False
    )
    assert [r.id for r in page1] == ["id0", "id1"]
    page2 = await storage.find_paginated_records_keyset(
        "config", last_id=2, limit=2, order_by="name", descending=False
    )
    assert [r.id for r in page2] == ["id2", "id3"]


@pytest.mark.asyncio
async def test_find_all_and_delete_all_and_search_session():
    from acapy_agent.storage.error import StorageSearchError
    from acapy_agent.storage.kanon_storage import KanonStorage, KanonStorageSearch
    from acapy_agent.storage.record import StorageRecord

    profile = FakeProfile()
    storage = KanonStorage(profile)

    for i in range(3):
        rec = StorageRecord(
            type="config",
            id=f"k{i}",
            value=f"v{i}",
            tags={"g": "1" if i % 2 == 0 else "2"},
        )
        await storage.add_record(rec)

    allrecs = await storage.find_all_records(
        "config", {"g": "1"}, order_by="name", descending=False
    )
    assert [r.id for r in allrecs] == ["k0", "k2"]

    search = KanonStorageSearch(profile)
    sess = search.search_records("config", {"g": "1"}, page_size=1)
    assert not sess.opened
    first = await sess.__anext__()
    assert first.id == "k0"
    got = await sess.fetch(max_count=2)
    if got:
        assert got[0].id == "k2"
    await sess.close()
    with pytest.raises(StorageSearchError):
        await sess.fetch(1)

    await storage.delete_all_records("config", {"g": "1"})
    remain = await storage.find_all_records("config", None)
    assert [r.id for r in remain] == ["k1"]


@pytest.mark.asyncio
async def test_storage_error_paths():
    from acapy_agent.storage.error import StorageDuplicateError, StorageNotFoundError
    from acapy_agent.storage.kanon_storage import KanonStorage
    from acapy_agent.storage.record import StorageRecord

    profile = FakeProfile()
    storage = KanonStorage(profile)

    rec = StorageRecord(type="t", id="a", value="v", tags={})
    await storage.add_record(rec)
    with pytest.raises(StorageDuplicateError):
        await storage.add_record(StorageRecord(type="t", id="a", value="v2", tags={}))

    with pytest.raises(StorageNotFoundError):
        await storage.get_record("t", "missing")

    with pytest.raises(StorageNotFoundError):
        await storage.update_record(
            StorageRecord(type="t", id="missing", value="v", tags={}), value="v2", tags={}
        )

    with pytest.raises(StorageNotFoundError):
        await storage.delete_record(
            StorageRecord(type="t", id="missing", value="v", tags={})
        )

    with pytest.raises(Exception):
        await storage.find_record("t", {"k": "v"})

    from acapy_agent.storage.error import StorageDuplicateError

    await storage.add_record(StorageRecord(type="t", id="b", value="v", tags={"k": "v"}))
    await storage.add_record(StorageRecord(type="t", id="c", value="v", tags={"k": "v"}))
    with pytest.raises(StorageDuplicateError):
        await storage.find_record("t", {"k": "v"})


@pytest.mark.asyncio
async def test_session_property_and_validations_and_error_mapping(monkeypatch):
    from acapy_agent.storage.error import StorageError
    from acapy_agent.storage.kanon_storage import KanonStorage
    from acapy_agent.storage.record import StorageRecord

    profile = FakeProfile()
    storage = KanonStorage(profile)

    assert storage.session is profile.dbstore_handle

    with pytest.raises(StorageError):
        await storage.get_record("", "id")
    with pytest.raises(StorageError):
        await storage.get_record("type", "")

    session = profile.session()
    rec = StorageRecord(type="x", id="one", value="v", tags={})
    await storage.add_record(rec, session=session)
    got = await storage.get_record("x", "one", session=session)
    assert got.id == "one"

    class BadSess(FakeStoreSession):
        async def fetch_all(self, *args, **kwargs):
            raise DBStoreError(DBStoreErrorCode.BUSY, "boom")

        async def __aenter__(self):
            return self

        async def __aexit__(self, et, ev, tb):
            return False

    def _bad_session():
        return BadSess(profile._handle)

    monkeypatch.setattr(profile, "session", _bad_session)
    with pytest.raises(StorageError):
        await storage.find_all_records("x", None)

    class BadDelSess(FakeStoreSession):
        async def remove_all(self, *args, **kwargs):
            raise DBStoreError(DBStoreErrorCode.BUSY, "fail")

        async def __aenter__(self):
            return self

        async def __aexit__(self, et, ev, tb):
            return False

    def _bad_del_session():
        return BadDelSess(profile._handle)

    monkeypatch.setattr(profile, "session", _bad_del_session)
    with pytest.raises(StorageError):
        await storage.delete_all_records("x", None)


@pytest.mark.asyncio
async def test_search_session_db_error(monkeypatch):
    from acapy_agent.storage.error import StorageSearchError
    from acapy_agent.storage.kanon_storage import KanonStorageSearch

    profile = FakeProfile()

    def _bad_scan(**kwargs):
        async def _gen():
            raise DBStoreError(DBStoreErrorCode.BUSY, "scan error")

        return _gen()

    monkeypatch.setattr(profile, "scan", _bad_scan)
    search = KanonStorageSearch(profile)
    sess = search.search_records("cat", {})
    with pytest.raises(StorageSearchError):
        await sess.__anext__()



"""Tests for credential definition creation with wait_for_revocation_setup options."""

import json
from unittest import IsolatedAsyncioTestCase

import pytest

from ....tests import mock
from ....utils.testing import create_test_profile
from ...issuer import AnonCredsIssuer
from ...models.credential_definition import CredDef
from ..revocation import AnonCredsRevocation
from ..revocation_setup import DefaultRevocationSetup


@pytest.mark.anoncreds
class TestAnonCredsIssuerWaitForRevocation(IsolatedAsyncioTestCase):
    """Tests for wait_for_revocation_setup functionality."""

    async def asyncSetUp(self) -> None:
        """Set up test environment."""
        self.profile = await create_test_profile(
            settings={"wallet.type": "askar-anoncreds"},
        )
        self.issuer = AnonCredsIssuer(self.profile)

    @mock.patch.object(AnonCredsIssuer, "notify")
    async def test_finish_cred_def_passes_options(self, mock_notify):
        """Test finish_cred_def method passes options correctly to the event."""
        # Mock transaction and entry data
        mock_entry = mock.MagicMock()
        mock_entry.value = json.dumps(
            {"issuer_id": "issuer-id", "schema_id": "schema-id"}
        )
        mock_entry.tags = {"support_revocation": "True", "max_cred_num": "1000"}

        self.profile.transaction = mock.Mock(
            return_value=mock.MagicMock(
                commit=mock.CoroutineMock(),
            )
        )

        with mock.patch.object(
            self.issuer, "_finish_registration", return_value=mock_entry
        ):
            with mock.patch.object(CredDef, "from_json") as mock_from_json:
                mock_cred_def = mock.MagicMock()
                mock_cred_def.schema_id = "schema-id"
                mock_cred_def.issuer_id = "issuer-id"
                mock_from_json.return_value = mock_cred_def

                await self.issuer.finish_cred_def(
                    job_id="job-id",
                    cred_def_id="cred-def-id",
                    options={"wait_for_revocation_setup": True},
                )

                # Should notify with correct parameters including options
                mock_notify.assert_called_once()
                call_args = mock_notify.call_args[0][0]  # Get the event passed to notify
                assert call_args.payload.cred_def_id == "cred-def-id"
                assert call_args.payload.support_revocation is True
                assert call_args.payload.options["wait_for_revocation_setup"] is True

    async def test_event_handler_respects_wait_option(self):
        """Test that the event handler respects the wait_for_revocation_setup option.

        This is a basic integration test to verify the event handler behavior.
        More comprehensive tests should be added to the revocation setup module.
        """
        # Create event handler
        setup_manager = DefaultRevocationSetup()

        # Create mock event with wait_for_revocation_setup=False
        mock_payload = mock.MagicMock()
        mock_payload.support_revocation = True
        mock_payload.cred_def_id = "test-cred-def-id"
        mock_payload.issuer_id = "test-issuer-id"
        mock_payload.max_cred_num = 1000
        mock_payload.options = {"wait_for_revocation_setup": False}

        event = mock.MagicMock()
        event.payload = mock_payload

        # Mock the AnonCredsRevocation class
        with mock.patch(
            "acapy_agent.anoncreds.revocation.revocation_setup.AnonCredsRevocation"
        ) as mock_revocation_class:
            mock_revocation = mock_revocation_class.return_value
            mock_revocation.emit_create_revocation_registry_definition_event = (
                mock.CoroutineMock()
            )
            mock_revocation.wait_for_active_revocation_registry = mock.CoroutineMock()

            # Call the event handler
            await setup_manager.on_cred_def(self.profile, event)

            # Should create registries but not wait
            mock_revocation.emit_create_revocation_registry_definition_event.assert_called_once()

            mock_revocation.wait_for_active_revocation_registry.assert_not_called()

    async def test_event_handler_waits_when_configured(self):
        """Test that the event handler waits when wait_for_revocation_setup=True."""
        # Create event handler
        setup_manager = DefaultRevocationSetup()

        # Create mock event with wait_for_revocation_setup=True
        mock_payload = mock.MagicMock()
        mock_payload.support_revocation = True
        mock_payload.cred_def_id = "test-cred-def-id"
        mock_payload.issuer_id = "test-issuer-id"
        mock_payload.max_cred_num = 1000
        mock_payload.options = {"wait_for_revocation_setup": True}

        event = mock.MagicMock()
        event.payload = mock_payload

        # Mock the AnonCredsRevocation class
        with mock.patch(
            "acapy_agent.anoncreds.revocation.revocation_setup.AnonCredsRevocation"
        ) as mock_revocation_class:
            mock_revocation = mock_revocation_class.return_value
            mock_revocation.emit_create_revocation_registry_definition_event = (
                mock.CoroutineMock()
            )
            mock_revocation.wait_for_active_revocation_registry = mock.CoroutineMock()

            # Call the event handler
            await setup_manager.on_cred_def(self.profile, event)

            # Should create registries AND wait
            mock_revocation.emit_create_revocation_registry_definition_event.assert_called_once()
            mock_revocation.wait_for_active_revocation_registry.assert_called_once_with(
                "test-cred-def-id"
            )


class TestAnonCredsRevocationWaitMethod(IsolatedAsyncioTestCase):
    """Test AnonCredsRevocation.wait_for_active_revocation_registry method."""

    async def asyncSetUp(self):
        """Set up test environment."""
        self.profile = await create_test_profile(
            settings={"wallet.type": "askar-anoncreds"}
        )
        self.revocation = AnonCredsRevocation(self.profile)
        self.cred_def_id = "test-cred-def-id"

    async def test_immediate_success_registry_already_active(self):
        """Test immediate success when registry is already active."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None  # Make sleep instant

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            mock_session.handle.fetch_all = mock.CoroutineMock(
                return_value=[{"id": "reg1"}]  # 1 active registry
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            with mock.patch.object(
                self.profile, "session", return_value=mock_session_context
            ):
                # Should complete immediately without timeout
                await self.revocation.wait_for_active_revocation_registry(
                    self.cred_def_id
                )

                # Should only query once
                mock_session.handle.fetch_all.assert_called_once_with(
                    "revocation_reg_def",
                    {"cred_def_id": self.cred_def_id, "active": "true"},
                )
                # Should not need to sleep
                mock_sleep.assert_not_called()

    async def test_success_after_polling(self):
        """Test successful completion after some polling iterations."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None  # Make sleep instant

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            # First 2 calls return empty, 3rd call returns 1 active registry
            mock_session.handle.fetch_all = mock.CoroutineMock(
                side_effect=[[], [], [{"id": "reg1"}]]
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            with mock.patch.object(
                self.profile, "session", return_value=mock_session_context
            ):
                await self.revocation.wait_for_active_revocation_registry(
                    self.cred_def_id
                )

                # Should have queried 3 times
                assert mock_session.handle.fetch_all.call_count == 3
                # Should have slept twice (after 1st and 2nd empty results)
                assert mock_sleep.call_count == 2
                mock_sleep.assert_called_with(0.5)

    async def test_timeout_no_active_registries(self):
        """Test timeout when no registries become active."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None  # Make sleep instant

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            mock_session.handle.fetch_all = mock.CoroutineMock(
                return_value=[]  # No active registries
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            # Set a very short timeout for testing
            with mock.patch(
                "acapy_agent.anoncreds.revocation.revocation.REVOCATION_REGISTRY_CREATION_TIMEOUT",
                1.0,
            ):
                with mock.patch.object(
                    self.profile, "session", return_value=mock_session_context
                ):
                    with self.assertRaises(TimeoutError) as exc_context:
                        await self.revocation.wait_for_active_revocation_registry(
                            self.cred_def_id
                        )

                    # Check error message content
                    error_message = str(exc_context.exception)
                    assert (
                        "Timeout waiting for revocation setup completion" in error_message
                    )
                    assert self.cred_def_id in error_message
                    assert "Expected 1 revocation registries" in error_message
                    assert "still be in progress in the background" in error_message

                    # Should have polled multiple times (1.0s timeout / 0.5s interval = 2 iterations)
                    assert mock_session.handle.fetch_all.call_count == 2

    async def test_polling_with_transient_errors_then_success(self):
        """Test that polling continues despite transient database errors."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None  # Make sleep instant

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            # Simulate: error, error, success
            mock_session.handle.fetch_all = mock.CoroutineMock(
                side_effect=[
                    Exception("Database connection error"),
                    Exception("Temporary network issue"),
                    [{"id": "reg1"}],  # Success on 3rd attempt
                ]
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            with mock.patch.object(
                self.profile, "session", return_value=mock_session_context
            ):
                await self.revocation.wait_for_active_revocation_registry(
                    self.cred_def_id
                )

                # Should have retried despite errors
                assert mock_session.handle.fetch_all.call_count == 3
                # Should have slept after each error
                assert mock_sleep.call_count == 2

    async def test_multiple_active_registries(self):
        """Test success when multiple registries are active (more than expected)."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None  # Make sleep instant

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            mock_session.handle.fetch_all = mock.CoroutineMock(
                return_value=[
                    {"id": "reg1"},
                    {"id": "reg2"},
                    {"id": "reg3"},
                ]  # 3 active registries
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            with mock.patch.object(
                self.profile, "session", return_value=mock_session_context
            ):
                await self.revocation.wait_for_active_revocation_registry(
                    self.cred_def_id
                )

                # Should complete immediately since we have >= 1
                mock_session.handle.fetch_all.assert_called_once()
                mock_sleep.assert_not_called()

    async def test_custom_timeout_value(self):
        """Test behavior with custom timeout configuration."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None  # Make sleep instant

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            mock_session.handle.fetch_all = mock.CoroutineMock(
                return_value=[]  # No active registries
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            # Set a custom timeout
            custom_timeout = 5.0
            with mock.patch(
                "acapy_agent.anoncreds.revocation.revocation.REVOCATION_REGISTRY_CREATION_TIMEOUT",
                custom_timeout,
            ):
                with mock.patch.object(
                    self.profile, "session", return_value=mock_session_context
                ):
                    with self.assertRaises(TimeoutError):
                        await self.revocation.wait_for_active_revocation_registry(
                            self.cred_def_id
                        )

                    # Should have polled based on custom timeout (5.0s / 0.5s = 10 iterations)
                    expected_iterations = int(custom_timeout / 0.5)
                    assert mock_session.handle.fetch_all.call_count == expected_iterations

    async def test_logging_behavior(self):
        """Test that appropriate log messages are generated."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            # First call empty, second call has 1 registry
            mock_session.handle.fetch_all = mock.CoroutineMock(
                side_effect=[[], [{"id": "reg1"}]]
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            with mock.patch(
                "acapy_agent.anoncreds.revocation.revocation.LOGGER"
            ) as mock_logger:
                with mock.patch.object(
                    self.profile, "session", return_value=mock_session_context
                ):
                    await self.revocation.wait_for_active_revocation_registry(
                        self.cred_def_id
                    )

                    # Should log debug message at start
                    mock_logger.debug.assert_any_call(
                        "Waiting for revocation setup completion for cred_def_id: %s",
                        self.cred_def_id,
                    )

                    # Should log progress updates
                    mock_logger.debug.assert_any_call(
                        "Revocation setup progress for %s: %d/%d registries active",
                        self.cred_def_id,
                        0,  # First iteration
                        1,  # Expected count
                    )

                    # Should log completion
                    mock_logger.info.assert_called_once_with(
                        "Revocation setup completed for cred_def_id: %s "
                        "(%d registries active)",
                        self.cred_def_id,
                        1,
                    )

    async def test_session_context_manager_usage(self):
        """Test that database session context manager is properly used."""
        with mock.patch("asyncio.sleep") as mock_sleep:
            mock_sleep.return_value = None

            # Mock the session and database query
            mock_session_context = mock.MagicMock()
            mock_session = mock.MagicMock()
            mock_session.handle.fetch_all = mock.CoroutineMock(
                return_value=[{"id": "reg1"}]  # Success immediately
            )
            mock_session_context.__aenter__ = mock.CoroutineMock(
                return_value=mock_session
            )
            mock_session_context.__aexit__ = mock.CoroutineMock(return_value=None)

            with mock.patch.object(
                self.profile, "session", return_value=mock_session_context
            ):
                await self.revocation.wait_for_active_revocation_registry(
                    self.cred_def_id
                )

                # Should have used the session context manager
                self.profile.session.assert_called_once()
                mock_session_context.__aenter__.assert_called_once()
                mock_session_context.__aexit__.assert_called_once()

                # Query should have been called with correct parameters
                mock_session.handle.fetch_all.assert_called_once_with(
                    "revocation_reg_def",
                    {"cred_def_id": self.cred_def_id, "active": "true"},
                )



"""Module docstring."""

CATEGORY = "revocation_list"

SCHEMAS = {
    "sqlite": [
        """
        CREATE TABLE IF NOT EXISTS revocation_list_v0_1 (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            item_id INTEGER NOT NULL,
            item_name TEXT NOT NULL,
            rev_reg_def_id TEXT,
            issuer_id TEXT,
            revocationList TEXT,  -- Note: json_valid() not available in older SQLite
            current_accumulator TEXT,
            next_index INTEGER NOT NULL DEFAULT 0,
            pending TEXT,
            state TEXT,
            rev_list TEXT,
            created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            updated_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            FOREIGN KEY (item_id) REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_item_id_v0_1 "
        "ON revocation_list_v0_1 (item_id);",
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_rev_reg_def_id_v0_1 "
        "ON revocation_list_v0_1 (rev_reg_def_id);",
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_issuer_id_v0_1 "
        "ON revocation_list_v0_1 (issuer_id);",
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_rev_reg_def_id_state_v0_1 "
        "ON revocation_list_v0_1 (rev_reg_def_id, state);",
        """
        CREATE TABLE IF NOT EXISTS revocation_list_revocations_v0_1 (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            revocation_list_id INTEGER NOT NULL,
            revoked_index INTEGER NOT NULL,
            FOREIGN KEY (revocation_list_id) REFERENCES revocation_list_v0_1(id)
            ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_revocations_revoked_index_v0_1 "
        "ON revocation_list_revocations_v0_1 (revoked_index);",
        """
        CREATE TRIGGER IF NOT EXISTS trg_insert_revocation_list_fields_v0_1
        AFTER INSERT ON revocation_list_v0_1
        FOR EACH ROW
        WHEN NEW.rev_list IS NOT NULL
        BEGIN
            UPDATE revocation_list_v0_1
            SET
                rev_reg_def_id = json_extract(NEW.rev_list, '$.revRegDefId'),
                issuer_id = json_extract(NEW.rev_list, '$.issuerId'),
                revocationList = json_extract(NEW.rev_list, '$.revocationList'),
                current_accumulator = json_extract(NEW.rev_list, '$.currentAccumulator')
            WHERE id = NEW.id;
        END;
        """,
        """
        CREATE TRIGGER IF NOT EXISTS trg_insert_revocation_list_revocations_v0_1
        AFTER INSERT ON revocation_list_v0_1
        FOR EACH ROW
        WHEN NEW.revocationList IS NOT NULL
        BEGIN
            INSERT INTO revocation_list_revocations_v0_1
            (revocation_list_id, revoked_index)
            SELECT NEW.id, key
            FROM json_each(NEW.revocationList)
            WHERE value = 1;
        END;
        """,
        """
        CREATE TRIGGER IF NOT EXISTS trg_update_revocation_list_fields_v0_1
        AFTER UPDATE ON revocation_list_v0_1
        FOR EACH ROW
        WHEN NEW.rev_list IS NOT NULL AND NEW.rev_list != OLD.rev_list
        BEGIN
            UPDATE revocation_list_v0_1
            SET
                rev_reg_def_id = json_extract(NEW.rev_list, '$.revRegDefId'),
                issuer_id = json_extract(NEW.rev_list, '$.issuerId'),
                revocationList = json_extract(NEW.rev_list, '$.revocationList'),
                current_accumulator = json_extract(NEW.rev_list, '$.currentAccumulator'),
                updated_at = CURRENT_TIMESTAMP
            WHERE id = NEW.id;
        END;
        """,
        """
        CREATE TRIGGER IF NOT EXISTS trg_update_revocation_list_revocations_v0_1
        AFTER UPDATE ON revocation_list_v0_1
        FOR EACH ROW
        WHEN NEW.revocationList IS NOT NULL AND NEW.revocationList != OLD.revocationList
        BEGIN
            DELETE FROM revocation_list_revocations_v0_1
            WHERE revocation_list_id = OLD.id;
            INSERT INTO revocation_list_revocations_v0_1
            (revocation_list_id, revoked_index)
            SELECT NEW.id, key
            FROM json_each(NEW.revocationList)
            WHERE value = 1;
        END;
        """,
        """
        CREATE TRIGGER IF NOT EXISTS trg_update_revocation_list_timestamp_v0_1
        AFTER UPDATE ON revocation_list_v0_1
        FOR EACH ROW
        BEGIN
            UPDATE revocation_list_v0_1
            SET updated_at = CURRENT_TIMESTAMP
            WHERE id = OLD.id;
        END;
        """,
    ],
    "postgresql": [
        """
        CREATE TABLE IF NOT EXISTS revocation_list_v0_1 (
            id SERIAL PRIMARY KEY,
            item_id INTEGER NOT NULL,
            item_name TEXT NOT NULL,
            rev_reg_def_id TEXT,
            issuer_id TEXT,
            revocationList TEXT CHECK (revocationList::jsonb IS NOT NULL),
            current_accumulator TEXT,
            next_index INTEGER NOT NULL DEFAULT 0,
            pending TEXT,
            state TEXT,
            rev_list TEXT,
            created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            CONSTRAINT fk_item_id FOREIGN KEY (item_id)
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_item_id_v0_1 "
        "ON revocation_list_v0_1 (item_id);",
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_rev_reg_def_id_v0_1 "
        "ON revocation_list_v0_1 (rev_reg_def_id);",
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_issuer_id_v0_1 "
        "ON revocation_list_v0_1 (issuer_id);",
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_rev_reg_def_id_state_v0_1 "
        "ON revocation_list_v0_1 (rev_reg_def_id, state);",
        """
        CREATE TABLE IF NOT EXISTS revocation_list_revocations_v0_1 (
            id SERIAL PRIMARY KEY,
            revocation_list_id INTEGER NOT NULL,
            revoked_index INTEGER NOT NULL,
            CONSTRAINT fk_revocation_list_id FOREIGN KEY (revocation_list_id)
            REFERENCES revocation_list_v0_1(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_revocation_list_revocations_revoked_index_v0_1 "
        "ON revocation_list_revocations_v0_1 (revoked_index);",
        """
        CREATE OR REPLACE FUNCTION insert_revocation_list_fields_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.rev_list IS NOT NULL AND NEW.rev_list::jsonb IS NOT NULL THEN
                NEW.rev_reg_def_id = jsonb_extract_path_text(
                    NEW.rev_list::jsonb, 'revRegDefId');
                NEW.issuer_id = jsonb_extract_path_text(NEW.rev_list::jsonb, 'issuerId');
                NEW.revocationList = jsonb_extract_path_text(
                    NEW.rev_list::jsonb, 'revocationList');
                NEW.current_accumulator = jsonb_extract_path_text(
                    NEW.rev_list::jsonb, 'currentAccumulator');
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_insert_revocation_list_fields_v0_1
        BEFORE INSERT ON revocation_list_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION insert_revocation_list_fields_v0_1();
        """,
        """
        CREATE OR REPLACE FUNCTION insert_revocation_list_revocations_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.revocationList IS NOT NULL 
            AND NEW.revocationList::jsonb IS NOT NULL THEN
                INSERT INTO revocation_list_revocations_v0_1
                (revocation_list_id, revoked_index)
                SELECT NEW.id, (key::INTEGER)
                FROM jsonb_array_elements(NEW.revocationList::jsonb)
                WITH ORDINALITY AS arr(value, key)
                WHERE value::INTEGER = 1;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_insert_revocation_list_revocations_v0_1
        AFTER INSERT ON revocation_list_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION insert_revocation_list_revocations_v0_1();
        """,
        """
        CREATE OR REPLACE FUNCTION update_revocation_list_fields_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.rev_list IS NOT NULL AND NEW.rev_list::jsonb IS NOT NULL 
               AND NEW.rev_list != OLD.rev_list THEN
                NEW.rev_reg_def_id = jsonb_extract_path_text(
                    NEW.rev_list::jsonb, 'revRegDefId');
                NEW.issuer_id = jsonb_extract_path_text(NEW.rev_list::jsonb, 'issuerId');
                NEW.revocationList = jsonb_extract_path_text(
                    NEW.rev_list::jsonb, 'revocationList');
                NEW.current_accumulator = jsonb_extract_path_text(
                    NEW.rev_list::jsonb, 'currentAccumulator');
                NEW.updated_at = CURRENT_TIMESTAMP;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_update_revocation_list_fields_v0_1
        BEFORE UPDATE ON revocation_list_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION update_revocation_list_fields_v0_1();
        """,
        """
        CREATE OR REPLACE FUNCTION update_revocation_list_revocations_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.revocationList IS NOT NULL AND NEW.revocationList::jsonb IS NOT NULL 
               AND NEW.revocationList != OLD.revocationList THEN
                DELETE FROM revocation_list_revocations_v0_1 
                WHERE revocation_list_id = OLD.id;
                INSERT INTO revocation_list_revocations_v0_1
                (revocation_list_id, revoked_index)
                SELECT NEW.id, (key::INTEGER)
                FROM jsonb_array_elements(NEW.revocationList::jsonb)
                WITH ORDINALITY AS arr(value, key)
                WHERE value::INTEGER = 1;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_update_revocation_list_revocations_v0_1
        AFTER UPDATE ON revocation_list_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION update_revocation_list_revocations_v0_1();
        """,
        """
        CREATE OR REPLACE FUNCTION update_revocation_list_timestamp_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.updated_at IS NULL THEN
                NEW.updated_at = CURRENT_TIMESTAMP;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_update_revocation_list_timestamp_v0_1
        BEFORE UPDATE ON revocation_list_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION update_revocation_list_timestamp_v0_1();
        """,
    ],
    "mssql": [
        """
        CREATE TABLE revocation_list_v0_1 (
            id INT IDENTITY(1,1) PRIMARY KEY,
            item_id INT NOT NULL,
            item_name NVARCHAR(MAX) NOT NULL,
            rev_reg_def_id NVARCHAR(255),
            issuer_id NVARCHAR(255),
            revocationList NVARCHAR(MAX) CHECK (ISJSON(revocationList) = 1),
            current_accumulator NVARCHAR(MAX),
            next_index INT NOT NULL DEFAULT 0,
            pending NVARCHAR(MAX),
            state NVARCHAR(255),
            rev_list NVARCHAR(MAX),
            created_at DATETIME2 DEFAULT SYSDATETIME(),
            updated_at DATETIME2 DEFAULT SYSDATETIME(),
            CONSTRAINT fk_item_id FOREIGN KEY (item_id)
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE NONCLUSTERED INDEX idx_revocation_list_item_id_v0_1 "
        "ON revocation_list_v0_1 (item_id);",
        "CREATE NONCLUSTERED INDEX idx_revocation_list_rev_reg_def_id_v0_1 "
        "ON revocation_list_v0_1 (rev_reg_def_id);",
        "CREATE NONCLUSTERED INDEX idx_revocation_list_issuer_id_v0_1 "
        "ON revocation_list_v0_1 (issuer_id);",
        "CREATE NONCLUSTERED INDEX idx_revocation_list_rev_reg_def_id_state_v0_1 "
        "ON revocation_list_v0_1 (rev_reg_def_id, state);",
        """
        CREATE TABLE revocation_list_revocations_v0_1 (
            id INT IDENTITY(1,1) PRIMARY KEY,
            revocation_list_id INT NOT NULL,
            revoked_index INT NOT NULL,
            CONSTRAINT fk_revocation_list_id FOREIGN KEY (revocation_list_id)
            REFERENCES revocation_list_v0_1(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE NONCLUSTERED INDEX idx_revocation_list_revocations_revoked_index_v0_1 "
        "ON revocation_list_revocations_v0_1 (revoked_index);",
        """
        CREATE TRIGGER trg_insert_revocation_list_fields_v0_1
        ON revocation_list_v0_1
        AFTER INSERT
        AS
        BEGIN
            UPDATE revocation_list_v0_1
            SET
                rev_reg_def_id = JSON_VALUE(i.rev_list, '$.revRegDefId'),
                issuer_id = JSON_VALUE(i.rev_list, '$.issuerId'),
                revocationList = JSON_VALUE(i.rev_list, '$.revocationList'),
                current_accumulator = JSON_VALUE(i.rev_list, '$.currentAccumulator')
            FROM revocation_list_v0_1 r
            INNER JOIN inserted i ON r.id = i.id
            WHERE i.rev_list IS NOT NULL AND ISJSON(i.rev_list) = 1;
        END;
        """,
        """
        CREATE TRIGGER trg_insert_revocation_list_revocations_v0_1
        ON revocation_list_v0_1
        AFTER INSERT
        AS
        BEGIN
            INSERT INTO revocation_list_revocations_v0_1
            (revocation_list_id, revoked_index)
            SELECT i.id, CAST(j.[key] AS INT)
            FROM inserted i
            CROSS APPLY OPENJSON(i.revocationList) j
            WHERE i.revocationList IS NOT NULL
              AND ISJSON(i.revocationList) = 1
              AND j.value = 1;
        END;
        """,
        """
        CREATE TRIGGER trg_update_revocation_list_fields_v0_1
        ON revocation_list_v0_1
        AFTER UPDATE
        AS
        BEGIN
            UPDATE revocation_list_v0_1
            SET
                rev_reg_def_id = JSON_VALUE(i.rev_list, '$.revRegDefId'),
                issuer_id = JSON_VALUE(i.rev_list, '$.issuerId'),
                revocationList = JSON_VALUE(i.rev_list, '$.revocationList'),
                current_accumulator = JSON_VALUE(i.rev_list, '$.currentAccumulator'),
                updated_at = SYSDATETIME()
            FROM revocation_list_v0_1 r
            INNER JOIN inserted i ON r.id = i.id
            WHERE i.rev_list IS NOT NULL
              AND ISJSON(i.rev_list) = 1
              AND i.rev_list != (SELECT d.rev_list FROM deleted d WHERE d.id = i.id);
        END;
        """,
        """
        CREATE TRIGGER trg_update_revocation_list_revocations_v0_1
        ON revocation_list_v0_1
        AFTER UPDATE
        AS
        BEGIN
            DELETE FROM revocation_list_revocations_v0_1
            WHERE revocation_list_id IN (
                SELECT i.id
                FROM inserted i
                INNER JOIN deleted d ON i.id = d.id
                WHERE i.revocationList IS NOT NULL
                  AND ISJSON(i.revocationList) = 1
                  AND i.revocationList != d.revocationList
            );

            INSERT INTO revocation_list_revocations_v0_1
            (revocation_list_id, revoked_index)
            SELECT i.id, CAST(j.[key] AS INT)
            FROM inserted i
            CROSS APPLY OPENJSON(i.revocationList) j
            WHERE i.revocationList IS NOT NULL
              AND ISJSON(i.revocationList) = 1
              AND j.value = 1
              AND i.revocationList != (
                SELECT d.revocationList FROM deleted d WHERE d.id = i.id
              );
        END;
        """,
        """
        CREATE TRIGGER trg_update_revocation_list_timestamp_v0_1
        ON revocation_list_v0_1
        AFTER UPDATE
        AS
        BEGIN
            UPDATE revocation_list_v0_1
            SET updated_at = SYSDATETIME()
            FROM revocation_list_v0_1
            INNER JOIN inserted ON revocation_list_v0_1.id = inserted.id
            WHERE inserted.updated_at IS NULL;
        END;
        """,
    ],
}


DROP_SCHEMAS = {
    "sqlite": [
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_timestamp_v0_1;",
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_revocations_v0_1;",
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_fields_v0_1;",
        "DROP TRIGGER IF EXISTS trg_insert_revocation_list_revocations_v0_1;",
        "DROP TRIGGER IF EXISTS trg_insert_revocation_list_fields_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_revocations_revoked_index_v0_1;",
        "DROP TABLE IF EXISTS revocation_list_revocations_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_rev_reg_def_id_state_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_issuer_id_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_rev_reg_def_id_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_item_id_v0_1;",
        "DROP TABLE IF EXISTS revocation_list_v0_1;",
    ],
    "postgresql": [
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_timestamp_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP FUNCTION IF EXISTS update_revocation_list_timestamp_v0_1 CASCADE;",
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_revocations_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP FUNCTION IF EXISTS update_revocation_list_revocations_v0_1 CASCADE;",
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_fields_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP FUNCTION IF EXISTS update_revocation_list_fields_v0_1 CASCADE;",
        "DROP TRIGGER IF EXISTS trg_insert_revocation_list_revocations_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP FUNCTION IF EXISTS insert_revocation_list_revocations_v0_1 CASCADE;",
        "DROP TRIGGER IF EXISTS trg_insert_revocation_list_fields_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP FUNCTION IF EXISTS insert_revocation_list_fields_v0_1 CASCADE;",
        "DROP INDEX IF EXISTS idx_revocation_list_revocations_revoked_index_v0_1;",
        "DROP TABLE IF EXISTS revocation_list_revocations_v0_1 CASCADE;",
        "DROP INDEX IF EXISTS idx_revocation_list_rev_reg_def_id_state_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_issuer_id_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_rev_reg_def_id_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_item_id_v0_1;",
        "DROP TABLE IF EXISTS revocation_list_v0_1 CASCADE;",
    ],
    "mssql": [
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_timestamp_v0_1;",
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_revocations_v0_1;",
        "DROP TRIGGER IF EXISTS trg_update_revocation_list_fields_v0_1;",
        "DROP TRIGGER IF EXISTS trg_insert_revocation_list_revocations_v0_1;",
        "DROP TRIGGER IF EXISTS trg_insert_revocation_list_fields_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_revocations_revoked_index_v0_1 "
        "ON revocation_list_revocations_v0_1;",
        "DROP TABLE IF EXISTS revocation_list_revocations_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_rev_reg_def_id_state_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_issuer_id_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_rev_reg_def_id_v0_1 "
        "ON revocation_list_v0_1;",
        "DROP INDEX IF EXISTS idx_revocation_list_item_id_v0_1 ON revocation_list_v0_1;",
        "DROP TABLE IF EXISTS revocation_list_v0_1;",
    ],
}

COLUMNS = ["rev_list", "next_index", "pending", "state"]



"""Module docstring."""

# handlers/custom/__init__.py
from .connection_metadata_custom_handler import ConnectionMetadataCustomHandler
from .cred_ex_v20_custom_handler import CredExV20CustomHandler
from .pres_ex_v20_custom_handler import PresExV20CustomHandler

__all__ = [
    "CredExV20CustomHandler",
    "ConnectionMetadataCustomHandler",
    "PresExV20CustomHandler",
]



"""Recover a revocation registry.

This module contains general exceptions or helper functions related to revocation
registry recovery that are not specific to any one implementation.
"""


class RevocRecoveryException(Exception):
    """Raise exception performing recovery."""



from unittest import IsolatedAsyncioTestCase

import pytest
from aiohttp import web
from aiohttp.web import HTTPForbidden, HTTPNotFound

from ......admin.request_context import AdminRequestContext
from ......tests import mock
from ......utils.testing import create_test_profile
from .....models.revocation import RevRegDef, RevRegDefValue
from .....revocation.revocation import AnonCredsRevocation
from .....tests.mock_objects import MockRevocationRegistryDefinition
from ....common.testing import BaseAnonCredsRouteTestCaseWithOutbound
from ..routes import get_tails_file, upload_tails_file


@pytest.mark.anoncreds
class TestAnonCredsTailsRoutes(
    BaseAnonCredsRouteTestCaseWithOutbound, IsolatedAsyncioTestCase
):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.rev_reg_id = (
            f"{self.test_did}:4:{self.test_did}:3:CL:1234:default:CL_ACCUM:default"
        )

    async def test_get_tails_file(self):
        self.request.match_info = {"rev_reg_id": self.rev_reg_id}

        with (
            mock.patch.object(
                AnonCredsRevocation,
                "get_created_revocation_registry_definition",
                mock.AsyncMock(),
            ) as mock_get_rev_reg,
            mock.patch.object(web, "FileResponse", mock.Mock()) as mock_file_response,
        ):
            mock_get_rev_reg.return_value = RevRegDef(
                issuer_id="issuer_id",
                type="CL_ACCUM",
                cred_def_id="cred_def_id",
                tag="tag",
                value=RevRegDefValue(
                    public_keys={},
                    max_cred_num=100,
                    tails_hash="tails_hash",
                    tails_location="tails_location",
                ),
            )

            result = await get_tails_file(self.request)
            mock_file_response.assert_called_once_with(path="tails_location", status=200)
            assert result is mock_file_response.return_value

    async def test_get_tails_file_not_found(self):
        self.request.match_info = {"rev_reg_id": self.rev_reg_id}

        with (
            mock.patch.object(
                AnonCredsRevocation,
                "get_created_revocation_registry_definition",
                mock.AsyncMock(),
            ) as mock_get_rev_reg,
            mock.patch.object(web, "FileResponse", mock.Mock()) as mock_file_response,
        ):
            mock_get_rev_reg.return_value = None

            with self.assertRaises(HTTPNotFound):
                await get_tails_file(self.request)
            mock_file_response.assert_not_called()

    async def test_tails_wrong_profile_403(self):
        """Test that tails file endpoints return 403 for wrong profile."""
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        self.request.match_info = {"rev_reg_id": "rev_reg_id"}
        with self.assertRaises(HTTPForbidden):
            await get_tails_file(self.request)

    @mock.patch.object(
        AnonCredsRevocation,
        "get_created_revocation_registry_definition",
        side_effect=[
            MockRevocationRegistryDefinition("revRegId"),
            None,
            MockRevocationRegistryDefinition("revRegId"),
        ],
    )
    @mock.patch.object(
        AnonCredsRevocation,
        "upload_tails_file",
        return_value=None,
    )
    async def test_upload_tails_file(self, mock_upload, mock_get):
        self.request.match_info = {"rev_reg_id": "rev_reg_id"}
        result = await upload_tails_file(self.request)
        assert result is not None
        assert mock_upload.call_count == 1
        assert mock_get.call_count == 1

        with self.assertRaises(HTTPNotFound):
            await upload_tails_file(self.request)

        self.request.match_info = {}

        with self.assertRaises(KeyError):
            await upload_tails_file(self.request)

    async def test_uploads_tails_wrong_profile_403(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        self.request.match_info = {"rev_reg_id": "rev_reg_id"}
        with self.assertRaises(HTTPForbidden):
            await upload_tails_file(self.request)



"""Kanon storage implementation for non-secrets storage."""

import asyncio
import inspect
import logging
from typing import Mapping, Optional, Sequence

from ..core.profile import Profile
from ..database_manager.dbstore import DBStoreError, DBStoreErrorCode, DBStoreSession
from .base import (
    DEFAULT_PAGE_SIZE,
    BaseStorage,
    BaseStorageSearch,
    BaseStorageSearchSession,
    validate_record,
)
from .error import (
    StorageDuplicateError,
    StorageError,
    StorageNotFoundError,
    StorageSearchError,
)
from .record import StorageRecord

LOGGER = logging.getLogger(__name__)

ERR_FETCH_SEARCH_RESULTS = "Error when fetching search results"


class KanonStorage(BaseStorage):
    """Kanon Non-Secrets interface."""

    def __init__(self, session: Profile):
        """Initialize KanonStorage with a profile session."""
        self._session = session

    @property
    def session(self) -> DBStoreSession:
        """Get the database session."""
        return self._session.dbstore_handle

    async def add_record(
        self, record: StorageRecord, session: Optional[DBStoreSession] = None
    ):
        """Add a new record to storage."""
        validate_record(record)
        if session is None:
            async with self._session.store.session() as temp_session:
                await self._add_record(record, temp_session)
        else:
            await self._add_record(record, session)

    async def _add_record(self, record: StorageRecord, session: DBStoreSession):
        try:
            await self._call_handle_or_session(
                session, "insert", record.type, record.id, record.value, record.tags
            )
        except DBStoreError as err:
            if err.code == DBStoreErrorCode.DUPLICATE:
                raise StorageDuplicateError(
                    f"Duplicate record: {record.type}/{record.id}"
                ) from None
            raise StorageError("Error when adding storage record") from err

    async def get_record(
        self,
        record_type: str,
        record_id: str,
        options: Optional[Mapping] = None,
        session: Optional[DBStoreSession] = None,
    ) -> StorageRecord:
        """Retrieve a single record by type and ID."""
        if not record_type:
            raise StorageError("Record type not provided")
        if not record_id:
            raise StorageError("Record ID not provided")
        for_update = bool(options and options.get("forUpdate"))
        if session is None:
            async with self._session.store.session() as temp_session:
                return await self._get_record(
                    record_type, record_id, for_update, temp_session
                )
        return await self._get_record(record_type, record_id, for_update, session)

    async def _get_record(
        self, record_type: str, record_id: str, for_update: bool, session: DBStoreSession
    ) -> StorageRecord:
        try:
            item = await self._call_handle_or_session(
                session, "fetch", record_type, record_id, for_update=for_update
            )
        except DBStoreError as err:
            raise StorageError("Error when fetching storage record") from err
        if not item:
            raise StorageNotFoundError(f"Record not found: {record_type}/{record_id}")
        return StorageRecord(
            type=item.category,
            id=item.name,
            value=item.value,
            tags=item.tags or {},
        )

    async def update_record(
        self,
        record: StorageRecord,
        value: str,
        tags: Mapping,
        session: Optional[DBStoreSession] = None,
    ):
        """Update an existing record's value and tags."""
        validate_record(record)
        if session is None:
            async with self._session.store.session() as temp_session:
                await self._update_record(record, value, tags, temp_session)
        else:
            await self._update_record(record, value, tags, session)

    async def _update_record(
        self, record: StorageRecord, value: str, tags: Mapping, session: DBStoreSession
    ):
        try:
            item = await self._call_handle_or_session(
                session, "fetch", record.type, record.id, for_update=True
            )
            if not item:
                raise StorageNotFoundError(f"Record not found: {record.type}/{record.id}")
            await self._call_handle_or_session(
                session, "replace", record.type, record.id, value, tags
            )
        except DBStoreError as err:
            if err.code == DBStoreErrorCode.NOT_FOUND:
                raise StorageNotFoundError(
                    f"Record not found: {record.type}/{record.id}"
                ) from None
            raise StorageError("Error when updating storage record value") from err

    async def delete_record(
        self, record: StorageRecord, session: Optional[DBStoreSession] = None
    ):
        """Delete a record from storage."""
        validate_record(record, delete=True)
        if session is None:
            async with self._session.store.session() as temp_session:
                await self._delete_record(record, temp_session)
        else:
            await self._delete_record(record, session)

    async def _delete_record(self, record: StorageRecord, session: DBStoreSession):
        try:
            await self._call_handle_or_session(session, "remove", record.type, record.id)
        except DBStoreError as err:
            if err.code == DBStoreErrorCode.NOT_FOUND:
                raise StorageNotFoundError(
                    f"Record not found: {record.type}/{record.id}"
                ) from None
            raise StorageError("Error when removing storage record") from err

    async def find_record(
        self,
        type_filter: str,
        tag_query: Mapping,
        options: Optional[Mapping] = None,
        session: Optional[DBStoreSession] = None,
    ) -> StorageRecord:
        """Find a single record matching the type and tag query."""
        for_update = bool(options and options.get("forUpdate"))
        if session is None:
            async with self._session.store.session() as temp_session:
                return await self._find_record(
                    type_filter, tag_query, for_update, temp_session
                )
        return await self._find_record(type_filter, tag_query, for_update, session)

    async def _find_record(
        self,
        type_filter: str,
        tag_query: Mapping,
        for_update: bool,
        session: DBStoreSession,
    ) -> StorageRecord:
        try:
            results = await self._call_handle_or_session(
                session,
                "fetch_all",
                type_filter,
                tag_query,
                limit=2,
                for_update=for_update,
            )
        except DBStoreError as err:
            raise StorageError("Error when finding storage record") from err
        if len(results) > 1:
            raise StorageDuplicateError("Duplicate records found")
        if not results:
            raise StorageNotFoundError("Record not found")
        row = results[0]
        return StorageRecord(
            type=row.category,
            id=row.name,
            value=row.value,
            tags=row.tags,
        )

    async def find_paginated_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        limit: int = DEFAULT_PAGE_SIZE,
        offset: int = 0,
        order_by: Optional[str] = None,
        descending: bool = False,
    ) -> Sequence[StorageRecord]:
        """Retrieve paginated records using DBStore.scan."""
        LOGGER.debug(
            "find_paginated_records: type=%s, tags=%s, limit=%s, "
            "offset=%s, order=%s, desc=%s",
            type_filter,
            tag_query,
            limit,
            offset,
            order_by,
            descending,
        )
        results = []
        scan = self._session.store.scan(
            category=type_filter,
            tag_filter=tag_query,
            limit=limit,
            offset=offset,
            profile=self._session.profile.name,
            order_by=order_by,
            descending=descending,
        )
        async for row in scan:
            results.append(
                StorageRecord(
                    type=row.category,
                    id=row.name,
                    value=row.value,
                    tags=row.tags,
                )
            )
        return results

    async def find_paginated_records_keyset(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        last_id: int = None,
        limit: int = DEFAULT_PAGE_SIZE,
        order_by: Optional[str] = None,
        descending: bool = False,
    ) -> Sequence[StorageRecord]:
        """Retrieve paginated records using DBStore.scan_keyset."""
        LOGGER.debug(
            "find_paginated_records_keyset: type=%s, tags=%s, last_id=%s, "
            "limit=%s, order=%s, desc=%s",
            type_filter,
            tag_query,
            last_id,
            limit,
            order_by,
            descending,
        )
        results = []
        scan = self._session.store.scan_keyset(
            category=type_filter,
            tag_filter=tag_query,
            last_id=last_id,
            limit=limit,
            profile=self._session.profile.name,
            order_by=order_by,
            descending=descending,
        )
        async for row in scan:
            results.append(
                StorageRecord(
                    type=row.category,
                    id=row.name,
                    value=row.value,
                    tags=row.tags,
                )
            )
        return results

    async def find_all_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        order_by: Optional[str] = None,
        descending: bool = False,
        options: Optional[Mapping] = None,
        session: Optional[DBStoreSession] = None,
    ) -> Sequence[StorageRecord]:
        """Retrieve all records matching the type and tag query."""
        for_update = bool(options and options.get("forUpdate"))
        if session is None:
            async with self._session.store.session() as temp_session:
                return await self._find_all_records(
                    type_filter, tag_query, order_by, descending, for_update, temp_session
                )
        return await self._find_all_records(
            type_filter, tag_query, order_by, descending, for_update, session
        )

    async def _find_all_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping],
        order_by: Optional[str],
        descending: bool,
        for_update: bool,
        session: DBStoreSession,
    ) -> Sequence[StorageRecord]:
        results = []
        try:
            for row in await self._call_handle_or_session(
                session,
                "fetch_all",
                type_filter,
                tag_query,
                order_by=order_by,
                descending=descending,
                for_update=for_update,
            ):
                results.append(
                    StorageRecord(
                        type=row.category,
                        id=row.name,
                        value=row.value,
                        tags=row.tags,
                    )
                )
        except DBStoreError as err:
            raise StorageError("Failed to fetch records") from err
        return results

    async def delete_all_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        session: Optional[DBStoreSession] = None,
    ):
        """Delete all records matching the type and tag query."""
        if session is None:
            async with self._session.store.session() as temp_session:
                await self._delete_all_records(type_filter, tag_query, temp_session)
        else:
            await self._delete_all_records(type_filter, tag_query, session)

    async def _delete_all_records(
        self, type_filter: str, tag_query: Optional[Mapping], session: DBStoreSession
    ):
        try:
            await self._call_handle_or_session(
                session, "remove_all", type_filter, tag_query
            )
        except DBStoreError as err:
            raise StorageError("Error when deleting records") from err

    async def _call_handle_or_session(self, session, method_name: str, *args, **kwargs):
        """Call a DB session method handling both sync handle.* and async session.*.

        If session.handle.<method> exists and is synchronous, call it directly.
        If it is asynchronous (coroutine or async generator),
        delegate to session.<method>.
        Otherwise, call/await session.<method> appropriately.
        """
        prefer_session_first = method_name in {"fetch_all", "remove_all"}

        if prefer_session_first:
            smethod = getattr(session, method_name, None)
            if smethod is not None and callable(smethod):
                try:
                    if inspect.iscoroutinefunction(smethod) or inspect.isasyncgenfunction(
                        smethod
                    ):
                        return await smethod(*args, **kwargs)
                    return smethod(*args, **kwargs)
                except TypeError:
                    handle = getattr(session, "handle", None)
                    if handle is not None and hasattr(handle, method_name):
                        hmethod = getattr(handle, method_name)
                        if inspect.iscoroutinefunction(hmethod):
                            return await hmethod(*args, **kwargs)
                        if inspect.isasyncgenfunction(hmethod):
                            results = []
                            async for item in hmethod(*args, **kwargs):
                                results.append(item)
                            return results
                        return hmethod(*args, **kwargs)
        else:
            handle = getattr(session, "handle", None)
            if handle is not None and hasattr(handle, method_name):
                hmethod = getattr(handle, method_name)
                if callable(hmethod):
                    if inspect.iscoroutinefunction(hmethod):
                        return await hmethod(*args, **kwargs)
                    if not inspect.isasyncgenfunction(hmethod):
                        return hmethod(*args, **kwargs)
            smethod = getattr(session, method_name, None)
            if smethod is not None and callable(smethod):
                if inspect.iscoroutinefunction(smethod) or inspect.isasyncgenfunction(
                    smethod
                ):
                    return await smethod(*args, **kwargs)
                return smethod(*args, **kwargs)

        handle = getattr(session, "handle", None)
        if handle is not None and hasattr(handle, method_name):
            hmethod = getattr(handle, method_name)
            if callable(hmethod):
                if inspect.iscoroutinefunction(hmethod):
                    return await hmethod(*args, **kwargs)
                if inspect.isasyncgenfunction(hmethod):
                    results = []
                    async for item in hmethod(*args, **kwargs):
                        results.append(item)
                    return results
                return hmethod(*args, **kwargs)
        raise AttributeError(f"Session does not provide method {method_name}")


class KanonStorageSearch(BaseStorageSearch):
    """Kanon storage search interface."""

    def __init__(self, profile: Profile):
        """Initialize KanonStorageSearch with a profile."""
        self._profile = profile

    def search_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        page_size: Optional[int] = None,
        options: Optional[Mapping] = None,
    ) -> "KanonStorageSearchSession":
        """Search for records."""
        return KanonStorageSearchSession(
            self._profile, type_filter, tag_query, page_size, options
        )


class KanonStorageSearchSession(BaseStorageSearchSession):
    """Kanon storage search session."""

    def __init__(
        self,
        profile,
        type_filter: str,
        tag_query: Mapping,
        page_size: Optional[int] = None,
        options: Optional[Mapping] = None,
    ):
        """Initialize search session with filter parameters."""
        self.tag_query = tag_query
        self.type_filter = type_filter
        self.page_size = page_size or DEFAULT_PAGE_SIZE
        self._done = False
        self._profile = profile
        self._scan = None
        self._timeout_task = None

    @property
    def opened(self) -> bool:
        """Check if search is opened."""
        return self._scan is not None

    @property
    def handle(self):
        """Get search handle."""
        return self._scan

    def __aiter__(self):
        """Return async iterator."""
        return self

    async def __anext__(self):
        """Get next item from search."""
        if self._done:
            raise StorageSearchError("Search query is complete")
        await self._open()
        try:
            if hasattr(self._scan, "__anext__"):
                row = await self._scan.__anext__()
            elif inspect.isawaitable(self._scan):
                # Awaitable scan: will raise DBStoreError per test, map to
                # StorageSearchError
                await self._scan
                await self.close()
                raise StopAsyncIteration
            else:
                # Synchronous iterator fallback
                row = next(self._scan)
            LOGGER.debug("Fetched row: category=%s, name=%s", row.category, row.name)
        except DBStoreError as err:
            await self.close()
            raise StorageSearchError(ERR_FETCH_SEARCH_RESULTS) from err
        except StopAsyncIteration:
            await self.close()
            raise
        return StorageRecord(
            type=row.category,
            id=row.name,
            value=row.value,  # DBStore returns a string from Entry.value
            tags=row.tags,
        )

    async def fetch(
        self, max_count: Optional[int] = None, offset: Optional[int] = None
    ) -> Sequence[StorageRecord]:
        """Fetch records."""
        if self._done:
            raise StorageSearchError("Search query is complete")
        limit = max_count or self.page_size
        await self._open(limit=limit, offset=offset)
        count = 0
        ret = []
        done = False
        if not hasattr(self._scan, "__anext__") and inspect.isawaitable(self._scan):
            try:
                await self._scan
            except DBStoreError as err:
                await self.close()
                raise StorageSearchError(ERR_FETCH_SEARCH_RESULTS) from err
            # No rows yielded
            await self.close()
            return ret
        while count < limit:
            try:
                if hasattr(self._scan, "__anext__"):
                    row = await self._scan.__anext__()
                else:
                    row = next(self._scan)
                LOGGER.debug("Fetched row: category=%s, name=%s", row.category, row.name)
                ret.append(
                    StorageRecord(
                        type=row.category,
                        id=row.name,
                        value=row.value,
                        tags=row.tags,
                    )
                )
                count += 1
            except DBStoreError as err:
                await self.close()
                raise StorageSearchError(ERR_FETCH_SEARCH_RESULTS) from err
            except StopAsyncIteration:
                done = True
                break
        if done or not ret:
            await self.close()
        return ret

    async def _open(self, offset: Optional[int] = None, limit: Optional[int] = None):
        if self._scan:
            return
        try:
            LOGGER.debug(
                "Opening scan for type_filter=%s, tag_query=%s, limit=%s, offset=%s",
                self.type_filter,
                self.tag_query,
                limit,
                offset,
            )
            self._scan = self._profile.opened.db_store.scan(
                category=self.type_filter,
                tag_filter=self.tag_query,
                offset=offset,
                limit=limit,
                profile=self._profile.name,
            )

            self._timeout_task = asyncio.create_task(self._timeout_close())
        except DBStoreError as err:
            raise StorageSearchError("Error opening search query") from err

    async def _timeout_close(self):
        """Close the scan after a timeout to prevent leaks."""
        await asyncio.sleep(30)
        if self._scan and not self._done:
            LOGGER.warning("Scan timeout reached, forcing closure")
            await self.close()

    async def close(self):
        """Close search session."""
        if self._timeout_task:
            self._timeout_task.cancel()
            self._timeout_task = None
        if self._scan:
            try:
                aclose = getattr(self._scan, "aclose", None)
                if aclose:
                    await aclose()
                else:
                    close = getattr(self._scan, "close", None)
                    if close:
                        res = close()
                        if inspect.iscoroutine(res):
                            await res
                LOGGER.debug("Closed KanonStorageSearchSession scan")
            except Exception:
                pass
            finally:
                self._scan = None
        self._done = True

    async def __aexit__(self, exc_type, exc, tb):
        """Exit async context manager."""
        await self.close()
        if exc_type:
            LOGGER.error("Exception in KanonStorageSearchSession: %s", exc)
        return False



"""PostgreSQL normalized database implementation."""

import asyncio
import logging
import threading
import time
import urllib.parse
from typing import TYPE_CHECKING, AsyncGenerator, Optional

from psycopg import pq
from psycopg_pool import AsyncConnectionPool

from ...category_registry import get_release
from ...db_types import Entry
from ...interfaces import AbstractDatabaseStore
from ...wql_normalized.query import query_from_str
from ...wql_normalized.tags import query_to_tagquery
from ..errors import DatabaseError, DatabaseErrorCode
from .connection_pool import PostgresConnectionPool
from .schema_context import SchemaContext

if TYPE_CHECKING:
    from .backend import PostgresqlBackend

LOGGER = logging.getLogger(__name__)


ERR_NO_DB_IN_CONN_STR = "No database name specified in connection string"


class PostgresDatabase(AbstractDatabaseStore):
    """PostgreSQL database implementation for normalized storage."""

    def __init__(
        self,
        pool: AsyncConnectionPool,
        default_profile: str,
        conn_str: str,
        release_number: str = "release_0",
        max_sessions: Optional[int] = None,
        min_size: Optional[int] = None,
        max_size: Optional[int] = None,
        timeout: Optional[float] = None,
        max_idle: Optional[float] = None,
        max_lifetime: Optional[float] = None,
        schema_context: Optional[SchemaContext] = None,
        backend: Optional["PostgresqlBackend"] = None,
    ):
        """Initialize PostgreSQL database."""
        self.lock = threading.RLock()
        self.pool = pool
        self.default_profile = default_profile
        self.conn_str = conn_str
        self.release_number = release_number
        self.active_sessions = []
        self.session_creation_times = {}
        self.max_sessions = (
            max_sessions if max_sessions is not None else int(pool.max_size * 0.75)
        )
        self.default_profile_id = None
        self.min_size = min_size if min_size is not None else pool.min_size
        self.max_size = max_size if max_size is not None else pool.max_size
        self.timeout = timeout if timeout is not None else pool.timeout
        self.max_idle = max_idle if max_idle is not None else pool.max_idle
        self.max_lifetime = (
            max_lifetime if max_lifetime is not None else pool.max_lifetime
        )
        self.schema_context = (
            schema_context or SchemaContext()
        )  # Default to SchemaContext
        self.backend = backend
        self._monitoring_task: Optional[asyncio.Task] = None

    async def initialize(self):
        """Initialize the database connection."""
        try:
            self.default_profile_id = await self._get_profile_id(self.default_profile)
        except Exception as e:
            LOGGER.error(
                "Failed to initialize default profile ID for '%s': %s",
                self.default_profile,
                str(e),
            )
            raise DatabaseError(
                code=DatabaseErrorCode.PROFILE_NOT_FOUND,
                message=(
                    f"Failed to initialize default profile ID for "
                    f"'{self.default_profile}'"
                ),
                actual_error=str(e),
            )

    async def start_monitoring(self):
        """Start monitoring active sessions."""
        if self._monitoring_task is None or self._monitoring_task.done():
            self._monitoring_task = asyncio.create_task(self._monitor_active_sessions())

    async def _monitor_active_sessions(self):
        while True:
            await asyncio.sleep(5)
            with self.lock:
                if self.active_sessions:
                    current_time = time.time()
                    for session in self.active_sessions[:]:
                        session_id = id(session)
                        creation_time = self.session_creation_times.get(session_id, 0)
                        age_seconds = current_time - creation_time
                        if age_seconds > 5:
                            try:
                                await session.close()
                            except Exception as e:
                                LOGGER.warning(
                                    "[monitor] Failed to close stale session %s: %s",
                                    id(session),
                                    str(e),
                                    exc_info=True,
                                )

    async def _get_profile_id(self, profile_name: str) -> int:
        conn = await self.pool.getconn()
        try:
            async with conn.cursor() as cursor:
                await cursor.execute(
                    (
                        f"SELECT id FROM "
                        f"{self.schema_context.qualify_table('profiles')} "
                        f"WHERE name = %s"
                    ),
                    (profile_name,),
                )
                row = await cursor.fetchone()
                if row:
                    return row[0]
                LOGGER.error("Profile '%s' not found", profile_name)
                raise DatabaseError(
                    code=DatabaseErrorCode.PROFILE_NOT_FOUND,
                    message=f"Profile '{profile_name}' not found",
                )
        except Exception as e:
            LOGGER.error(
                "Failed to retrieve profile ID for '%s': %s", profile_name, str(e)
            )
            raise DatabaseError(
                code=DatabaseErrorCode.QUERY_ERROR,
                message=f"Failed to retrieve profile ID for '{profile_name}'",
                actual_error=str(e),
            )
        finally:
            await self.pool.putconn(conn)

    async def create_profile(self, name: str = None) -> str:
        """Create a new profile."""
        name = name or "new_profile"
        conn = await self.pool.getconn()
        try:
            async with conn.cursor() as cursor:
                await cursor.execute(
                    (
                        f"INSERT INTO "
                        f"{self.schema_context.qualify_table('profiles')} "
                        f"(name, profile_key) VALUES (%s, NULL) "
                        f"ON CONFLICT (name) DO NOTHING"
                    ),
                    (name,),
                )
                if cursor.rowcount == 0:
                    LOGGER.error("Profile '%s' already exists", name)
                    raise DatabaseError(
                        code=DatabaseErrorCode.PROFILE_ALREADY_EXISTS,
                        message=f"Profile '{name}' already exists",
                    )
                await conn.commit()
                return name
        except Exception as e:
            await conn.rollback()
            LOGGER.error("Failed to create profile '%s': %s", name, str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.QUERY_ERROR,
                message=f"Failed to create profile '{name}'",
                actual_error=str(e),
            )
        finally:
            await self.pool.putconn(conn)

    async def get_profile_name(self) -> str:
        """Get the default profile name."""
        return self.default_profile

    async def remove_profile(self, name: str) -> bool:
        """Remove a profile."""
        conn = await self.pool.getconn()
        try:
            async with conn.cursor() as cursor:
                await cursor.execute(
                    (
                        f"DELETE FROM "
                        f"{self.schema_context.qualify_table('profiles')} "
                        f"WHERE name = %s"
                    ),
                    (name,),
                )
                result = cursor.rowcount > 0
                await conn.commit()
                return result
        except Exception as e:
            await conn.rollback()
            LOGGER.error("Failed to remove profile '%s': %s", name, str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.QUERY_ERROR,
                message=f"Failed to remove profile '{name}'",
                actual_error=str(e),
            )
        finally:
            await self.pool.putconn(conn)

    async def rekey(self, key_method: str = None, pass_key: str = None):
        """Rekey the database (not supported for PostgreSQL)."""
        LOGGER.error("Rekey not supported for PostgreSQL")
        raise DatabaseError(
            code=DatabaseErrorCode.UNSUPPORTED_OPERATION,
            message="Rekey not supported for PostgreSQL",
        )

    async def scan(
        self,
        profile: Optional[str],
        category: str,
        tag_filter: str | dict = None,
        offset: int = None,
        limit: int = None,
        order_by: Optional[str] = None,
        descending: bool = False,
    ) -> AsyncGenerator[Entry, None]:
        """Scan for entries matching criteria."""
        handlers, _, _ = get_release(self.release_number, "postgresql")

        handler = handlers.get(category, handlers["default"])
        # Update handler's schema_context to match database's schema_context
        if hasattr(handler, "set_schema_context"):
            handler.set_schema_context(self.schema_context)
        profile_id = await self._get_profile_id(profile or self.default_profile)
        tag_query = None
        if tag_filter:
            wql_query = query_from_str(tag_filter)
            tag_query = query_to_tagquery(wql_query)
        conn = await self.pool.getconn()
        try:
            async with conn.cursor() as cursor:
                async for entry in handler.scan(
                    cursor,
                    profile_id,
                    category,
                    tag_query,
                    offset,
                    limit,
                    order_by,
                    descending,
                ):
                    yield entry
                if conn.pgconn.transaction_status != pq.TransactionStatus.IDLE:
                    await conn.commit()
        except Exception as e:
            await conn.rollback()
            error_message = f"Failed to execute scan query: {str(e)}"
            LOGGER.error(error_message)
            raise DatabaseError(
                code=DatabaseErrorCode.QUERY_ERROR,
                message=error_message,
                actual_error=str(e),
            )
        finally:
            await self.pool.putconn(conn)

    async def scan_keyset(
        self,
        profile: Optional[str],
        category: str,
        tag_filter: str | dict = None,
        last_id: Optional[int] = None,
        limit: int = None,
        order_by: Optional[str] = None,
        descending: bool = False,
    ) -> AsyncGenerator[Entry, None]:
        """Scan using keyset pagination."""
        handlers, _, _ = get_release(self.release_number, "postgresql")

        handler = handlers.get(category, handlers["default"])
        # Update handler's schema_context to match database's schema_context
        if hasattr(handler, "set_schema_context"):
            handler.set_schema_context(self.schema_context)
        profile_id = await self._get_profile_id(profile or self.default_profile)
        tag_query = None
        if tag_filter:
            wql_query = query_from_str(tag_filter)
            tag_query = query_to_tagquery(wql_query)
        conn = await self.pool.getconn()
        try:
            async with conn.cursor() as cursor:
                async for entry in handler.scan_keyset(
                    cursor,
                    profile_id,
                    category,
                    tag_query,
                    last_id,
                    limit,
                    order_by,
                    descending,
                ):
                    yield entry
                if conn.pgconn.transaction_status != pq.TransactionStatus.IDLE:
                    await conn.commit()
        except Exception as e:
            await conn.rollback()
            error_message = f"Failed to execute scan_keyset query: {str(e)}"
            LOGGER.error(error_message)
            raise DatabaseError(
                code=DatabaseErrorCode.QUERY_ERROR,
                message=error_message,
                actual_error=str(e),
            )
        finally:
            await self.pool.putconn(conn)

    async def session(self, profile: str = None):
        """Create a new database session."""
        from .session import PostgresSession

        with self.lock:
            if len(self.active_sessions) >= self.max_sessions:
                LOGGER.error(
                    "Maximum number of active sessions reached: %d", self.max_sessions
                )
                raise DatabaseError(
                    code=DatabaseErrorCode.CONNECTION_POOL_EXHAUSTED,
                    message="Maximum number of active sessions reached",
                )
        effective_profile = profile or self.default_profile
        cached_profile_id = (
            self.default_profile_id if effective_profile == self.default_profile else None
        )
        sess = PostgresSession(
            self, effective_profile, False, self.release_number, cached_profile_id
        )
        with self.lock:
            self.active_sessions.append(sess)
            self.session_creation_times[id(sess)] = time.time()
        LOGGER.debug(
            "[session] Active sessions: %d, session_id=%s",
            len(self.active_sessions),
            id(sess),
        )
        return sess

    async def transaction(self, profile: str = None):
        """Create a new database transaction."""
        from .session import PostgresSession

        with self.lock:
            if len(self.active_sessions) >= self.max_sessions:
                LOGGER.error(
                    "Maximum number of active sessions reached: %d", self.max_sessions
                )
                raise DatabaseError(
                    code=DatabaseErrorCode.CONNECTION_POOL_EXHAUSTED,
                    message="Maximum number of active sessions reached",
                )
        effective_profile = profile or self.default_profile
        cached_profile_id = (
            self.default_profile_id if effective_profile == self.default_profile else None
        )
        sess = PostgresSession(
            self, effective_profile, True, self.release_number, cached_profile_id
        )
        with self.lock:
            self.active_sessions.append(sess)
            self.session_creation_times[id(sess)] = time.time()
        LOGGER.debug(
            "[session] Active sessions: %d, session_id=%s",
            len(self.active_sessions),
            id(sess),
        )
        return sess

    async def close(self, remove: bool = False):
        """Close the database connection."""
        try:
            # Cancel background monitoring task if running
            if self._monitoring_task and not self._monitoring_task.done():
                self._monitoring_task.cancel()
                try:
                    await self._monitoring_task
                finally:
                    self._monitoring_task = None
            if remove:
                parsed = urllib.parse.urlparse(self.conn_str)
                target_db = parsed.path.lstrip("/")
                if not target_db:
                    raise ValueError(ERR_NO_DB_IN_CONN_STR)
                default_conn_str = self.conn_str.replace(f"/{target_db}", "/postgres")
                pool = PostgresConnectionPool(
                    conn_str=default_conn_str,
                    min_size=self.min_size,
                    max_size=self.max_size,
                    timeout=self.timeout,
                    max_idle=self.max_idle,
                    max_lifetime=self.max_lifetime,
                )
                await pool.initialize()
                try:
                    conn = await pool.getconn()
                    try:
                        await conn.rollback()
                        await conn.set_autocommit(True)
                        async with conn.cursor() as cursor:
                            await cursor.execute(
                                (
                                    "SELECT pg_terminate_backend(pg_stat_activity.pid) "
                                    "FROM pg_stat_activity "
                                    "WHERE pg_stat_activity.datname = %s "
                                    "AND pid <> pg_backend_pid()"
                                ),
                                (target_db,),
                            )
                            await cursor.execute(f"DROP DATABASE IF EXISTS {target_db}")
                    except Exception as e:
                        LOGGER.error("Failed to drop database %s: %s", target_db, str(e))
                        raise DatabaseError(
                            code=DatabaseErrorCode.CONNECTION_ERROR,
                            message=f"Failed to drop database {target_db}",
                            actual_error=str(e),
                        )
                    finally:
                        await pool.putconn(conn)
                finally:
                    await pool.close()
            await self.pool.close()
        except Exception as e:
            LOGGER.error("Failed to close database: %s", str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.CONNECTION_ERROR,
                message="Failed to close database",
                actual_error=str(e),
            )



"""AnonCreds revocation routes."""

from ....messaging.models.openapi import OpenAPISchema

REVOCATION_TAG_TITLE = "AnonCreds - Revocation"


class AnonCredsRevocationModuleResponseSchema(OpenAPISchema):
    """Response schema for Revocation Module."""


__all__ = ["REVOCATION_TAG_TITLE", "AnonCredsRevocationModuleResponseSchema"]



"""Tests for the tags module.

Disabled by default to keep CI lean; enable locally with
ENABLE_WQL_SQLITE_TESTS=1 if you want to run them.
"""

import os

import pytest

if not os.getenv("ENABLE_WQL_SQLITE_TESTS"):
    pytest.skip(
        "WQL SQLite encoder tests disabled by default; set ENABLE_WQL_SQLITE_TESTS=1",
        allow_module_level=True,
    )

import unittest
from typing import List

from ..query import AndQuery, EqQuery
from ..tags import CompareOp, ConjunctionOp, TagName, TagQuery, query_to_tagquery


class TestEncoder:
    """A class to encode TagQuery objects into string representations."""

    def encode_query(self, query: TagQuery, negate: bool = False) -> str:
        """Encode a TagQuery into a string representation, handling negation."""
        if query.variant in ["Eq", "Neq", "Gt", "Gte", "Lt", "Lte", "Like"]:
            op = getattr(CompareOp, query.variant)
            return self.encode_op(op, *query.data, negate)
        elif query.variant == "In":
            return self.encode_in(*query.data, negate)
        elif query.variant == "Exist":
            return self.encode_exist(query.data, negate)
        elif query.variant in ["And", "Or"]:
            op = ConjunctionOp.And if query.variant == "And" else ConjunctionOp.Or
            return self.encode_conj(op, query.data, negate)
        elif query.variant == "Not":
            return self.encode_query(query.data, not negate)
        else:
            raise ValueError(f"Unknown query variant: {query.variant}")

    def encode_op(self, op: CompareOp, name: TagName, value: str, negate: bool) -> str:
        """Encode an operation clause (e.g., 'name = value')."""
        enc_name = self.encode_name(name)
        enc_value = self.encode_value(value)
        clause = f"{enc_name} {op.as_sql_str()} {enc_value}"
        return f"NOT ({clause})" if negate else clause

    def encode_in(self, name: TagName, values: List[str], negate: bool) -> str:
        """Encode an IN clause (e.g., 'name IN (value1, value2)')."""
        enc_name = self.encode_name(name)
        enc_values = [self.encode_value(v) for v in values]
        op_str = "NOT IN" if negate else "IN"
        values_str = ", ".join(enc_values)
        return f"{enc_name} {op_str} ({values_str})"

    def encode_exist(self, names: List[TagName], negate: bool) -> str:
        """Encode an EXIST clause (e.g., 'EXIST(name)')."""

        if not names:
            return None
        clauses = []
        for name in names:
            enc_name = self.encode_name(name)
            op_str = "NOT EXIST" if negate else "EXIST"
            clauses.append(f"{op_str}({enc_name})")
        if len(clauses) == 1:
            return clauses[0]
        op = ConjunctionOp.And if not negate else ConjunctionOp.Or
        return f"({op.as_sql_str().join(clauses)})"

    def encode_conj(
        self, op: ConjunctionOp, subqueries: List[TagQuery], negate: bool
    ) -> str:
        """Encode a conjunction clause (AND/OR) with possible negation."""
        if negate:
            op = op.negate()
            sub_negate = True
        else:
            sub_negate = False
        clauses = [self.encode_query(q, sub_negate) for q in subqueries if q is not None]
        if not clauses:
            return None
        return f"({op.as_sql_str().join(clauses)})"

    def encode_name(self, name: TagName) -> str:
        """Test cases encode name functionality."""

        return name.to_string()

    def encode_value(self, value: str) -> str:
        """Test cases for the TagQuery functionality."""

        return value


class TestTags(unittest.TestCase):
    """Test cases for the TagQuery functionality."""

    def test_from_query(self):
        """Test cases for the TagQuery functionality."""
        query = AndQuery([EqQuery("enctag", "encval"), EqQuery("~plaintag", "plainval")])
        tag_query = query_to_tagquery(query)
        self.assertEqual(tag_query.variant, "And")
        self.assertEqual(len(tag_query.data), 2)
        sq1, sq2 = tag_query.data
        self.assertEqual(sq1.variant, "Eq")
        name1, val1 = sq1.data
        self.assertEqual(name1.value, "enctag")
        self.assertEqual(val1, "encval")
        self.assertEqual(sq2.variant, "Eq")
        name2, val2 = sq2.data
        self.assertEqual(name2.value, "plaintag")
        self.assertEqual(val2, "plainval")

    def test_serialize(self):
        """Test serialization of TagQuery to JSON."""
        self.skipTest("TagQuery serialization not implemented in provided code")

    def test_simple_and(self):
        """Test encoding a complex TagQuery with AND, OR, and NOT."""
        condition_1 = TagQuery.and_(
            [
                TagQuery.eq(TagName("enctag"), "encval"),
                TagQuery.eq(TagName("plaintag"), "plainval"),
            ]
        )
        condition_2 = TagQuery.and_(
            [
                TagQuery.eq(TagName("enctag"), "encval"),
                TagQuery.not_(TagQuery.eq(TagName("plaintag"), "eggs")),
            ]
        )
        query = TagQuery.or_([condition_1, condition_2])
        encoder = TestEncoder()
        query_str = encoder.encode_query(query)
        expected = (
            "((enctag = encval AND plaintag = plainval) OR "
            "(enctag = encval AND NOT (plaintag = eggs)))"
        )
        self.assertEqual(query_str, expected)

    def test_negate_conj(self):
        """Test encoding a negated conjunction TagQuery."""
        condition_1 = TagQuery.and_(
            [
                TagQuery.eq(TagName("enctag"), "encval"),
                TagQuery.eq(TagName("plaintag"), "plainval"),
            ]
        )
        condition_2 = TagQuery.and_(
            [
                TagQuery.eq(TagName("enctag"), "encval"),
                TagQuery.not_(TagQuery.eq(TagName("plaintag"), "eggs")),
            ]
        )
        query = TagQuery.not_(TagQuery.or_([condition_1, condition_2]))
        encoder = TestEncoder()
        query_str = encoder.encode_query(query)
        expected = (
            "((NOT (enctag = encval) OR NOT (plaintag = plainval)) AND "
            "(NOT (enctag = encval) OR plaintag = eggs))"
        )
        self.assertEqual(query_str, expected)


if __name__ == "__main__":
    unittest.main()



import json
import types
from typing import Any, Dict, Optional

import pytest


@pytest.fixture
def wallet_env(monkeypatch):
    from acapy_agent.wallet import kanon_wallet as module

    class FakeKeyAlg:
        def __init__(self, value: str):
            self.value = value

    class FakeKey:
        _seq = 0

        def __init__(self, alg: Any, pb: bytes, sb: bytes):
            self.algorithm = alg
            self._pb = pb
            self._sb = sb

        @staticmethod
        def generate(alg):
            FakeKey._seq += 1
            return FakeKey(
                FakeKeyAlg(getattr(alg, "value", str(alg))),
                f"pub{FakeKey._seq}".encode(),
                b"sec",
            )

        def get_public_bytes(self):
            return self._pb

        def get_secret_bytes(self):
            return self._sb

        def sign_message(self, m):
            return b"sig" + (m if isinstance(m, bytes) else b"".join(m))

    class _KeyAlg:
        ED25519 = FakeKeyAlg("ed25519")
        X25519 = FakeKeyAlg("x25519")
        P256 = FakeKeyAlg("p256")
        BLS12_381_G2 = FakeKeyAlg("bls12_381_g2")

    monkeypatch.setattr(module, "Key", FakeKey, raising=True)
    monkeypatch.setattr(module, "KeyAlg", _KeyAlg, raising=True)
    monkeypatch.setattr(module, "validate_seed", lambda s: b"seedbytes")

    class FakeAskar:
        def __init__(self):
            self._keys: Dict[str, Dict[str, Any]] = {}

        async def insert_key(
            self,
            name: str,
            key: FakeKey,
            metadata: Optional[str] = None,
            tags: Optional[dict] = None,
        ):
            if name in self._keys:

                class _Err(Exception):
                    def __init__(self):
                        self.code = "DUPLICATE"

                raise _Err()
            self._keys[name] = {"key": key, "metadata": metadata, "tags": tags or {}}

        async def fetch_key(self, name: str, for_update: bool = False):
            entry = self._keys.get(name)
            if not entry:
                return None
            return types.SimpleNamespace(
                key=entry["key"], metadata=entry["metadata"], tags=entry["tags"]
            )

        async def update_key(
            self, name: str, tags: Optional[dict] = None, metadata: Optional[str] = None
        ):
            entry = self._keys.get(name)
            if not entry:

                class _Err(Exception):
                    def __init__(self):
                        self.code = "NOT_FOUND"

                raise _Err()
            if tags is not None:
                entry["tags"] = tags
            if metadata is not None:
                entry["metadata"] = metadata

        async def fetch_all_keys(self, tag_filter: dict, limit: int = 2):
            results = []
            for _, entry in self._keys.items():
                if all(entry["tags"].get(k) == v for k, v in (tag_filter or {}).items()):
                    results.append(
                        types.SimpleNamespace(
                            key=entry["key"],
                            metadata=entry["metadata"],
                            tags=entry["tags"],
                        )
                    )
                    if len(results) >= limit:
                        break
            return results

    class FakeStore:
        def __init__(self):
            self._rows: Dict[tuple[str, str], Dict[str, Any]] = {}

        def session(self):
            store = self

            class Sess:
                async def __aenter__(self):
                    return self

                async def __aexit__(self, et, ev, tb):
                    return False

                async def fetch(self, cat, name, for_update: bool = False):
                    row = store._rows.get((cat, name))
                    if not row:
                        return None
                    return types.SimpleNamespace(
                        category=cat,
                        name=name,
                        value=row.get("value"),
                        value_json=row.get("value_json"),
                        tags=row.get("tags", {}),
                    )

                async def replace(
                    self, cat, name, value=None, tags=None, value_json=None
                ):
                    row = store._rows.get((cat, name))
                    if not row:
                        return None
                    if value_json is not None:
                        row["value_json"] = value_json
                        row["value"] = json.dumps(value_json)
                    if value is not None:
                        row["value"] = value
                    if tags is not None:
                        row["tags"] = tags

                async def insert(self, cat, name, value=None, tags=None, value_json=None):
                    store._rows[(cat, name)] = {
                        "value": value,
                        "value_json": value_json,
                        "tags": tags or {},
                    }

                async def fetch_all(self, cat, tag_filter=None, **kwargs):
                    res = []
                    for (c, n), row in store._rows.items():
                        if c != cat:
                            continue
                        if tag_filter and any(
                            (row.get("tags", {}).get(k) != v)
                            for k, v in tag_filter.items()
                        ):
                            continue
                        res.append(
                            types.SimpleNamespace(
                                category=c,
                                name=n,
                                value=row.get("value"),
                                value_json=row.get("value_json"),
                                tags=row.get("tags", {}),
                            )
                        )
                    return res

            return Sess()

        async def scan(self, **kwargs):
            if False:
                yield None

    class FakeProfile:
        def __init__(self):
            self.askar_handle = FakeAskar()
            self.store = FakeStore()
            self.context = types.SimpleNamespace(inject=lambda c: None)
            self.profile = types.SimpleNamespace(name="p")

    key_types = module.KeyTypes()

    class _DIDMethods:
        def from_method(self, name):
            return module.SOV

        def from_did(self, did):
            return types.SimpleNamespace(method_name="sov", supports_rotation=True)

    class _FakeDPV:
        def __init__(self, *_):
            pass

        def validate_key_type(self, *_):
            return None

        def validate_or_derive_did(self, method, key_type, verkey_bytes, did):
            return did or "did:sov:testdid"

    monkeypatch.setattr(module, "DIDParametersValidation", _FakeDPV, raising=True)

    profile = FakeProfile()

    class _Session:
        def __init__(self, profile):
            self.askar_handle = profile.askar_handle
            self.store = profile.store
            self.context = types.SimpleNamespace(
                inject=lambda cls: key_types if cls is module.KeyTypes else _DIDMethods()
            )
            self.profile = profile

        def inject(self, cls):
            return key_types if cls is module.KeyTypes else _DIDMethods()

    session = _Session(profile)

    wallet = module.KanonWallet(session)
    return module, wallet, profile


@pytest.mark.asyncio
async def test_input_and_lookup_errors(wallet_env):
    module, wallet, profile = wallet_env

    with pytest.raises(module.WalletNotFoundError):
        await wallet.get_signing_key("")

    with pytest.raises(module.WalletNotFoundError):
        await wallet.replace_signing_key_metadata("nope", {})

    with pytest.raises(module.WalletNotFoundError):
        await wallet.get_key_by_kid("kid-missing")

    info1 = await wallet.create_key(module.ED25519, metadata={}, kid="dup")
    k2 = module.Key.generate(module.ED25519)
    await profile.askar_handle.insert_key(
        name="v2", key=k2, metadata=json.dumps({}), tags={"kid": "dup"}
    )
    with pytest.raises(module.WalletDuplicateError):
        await wallet.get_key_by_kid("dup")

    with pytest.raises(module.WalletError):
        await wallet.sign_message(b"", info1.verkey)
    with pytest.raises(module.WalletError):
        await wallet.sign_message(b"msg", "")
    with pytest.raises(module.WalletError):
        await wallet.verify_message(b"msg", b"", info1.verkey, module.ED25519)
    with pytest.raises(module.WalletError):
        await wallet.verify_message(b"", b"sig", info1.verkey, module.ED25519)


@pytest.mark.asyncio
async def test_set_did_endpoint_and_ledger_errors(wallet_env):
    module, wallet, profile = wallet_env

    did_info = await wallet.create_local_did(
        module.SOV, module.ED25519, metadata={"public": True}
    )
    await wallet.set_public_did(did_info.did)

    with pytest.raises(module.LedgerConfigError):
        await wallet.set_did_endpoint(did_info.did, "http://e", ledger=None)

    class FakeLedger:
        def __init__(self):
            self.read_only = False

        async def __aenter__(self):
            return self

        async def __aexit__(self, et, ev, tb):
            return False

        async def update_endpoint_for_did(
            self,
            did,
            endpoint,
            endpoint_type,
            write_ledger=True,
            endorser_did=None,
            routing_keys=None,
        ):
            return {"did": did, "endpoint": endpoint}

    attrib = await wallet.set_did_endpoint(
        did_info.did, "http://e", ledger=FakeLedger(), write_ledger=False
    )
    assert attrib["did"] == did_info.did


@pytest.mark.asyncio
async def test_assign_kid_and_get_by_kid_success(wallet_env, monkeypatch):
    module, wallet, profile = wallet_env

    info = await wallet.create_key(module.ED25519)
    assigned = await wallet.assign_kid_to_key(info.verkey, "kid-ok")
    assert assigned.kid == "kid-ok"

    got = await wallet.get_key_by_kid("kid-ok")
    assert got.verkey == info.verkey and got.kid == "kid-ok"

    class BadKeyTypes(module.KeyTypes.__class__):
        def from_key_type(self, *_):
            return None

    monkeypatch.setattr(
        wallet.session,
        "inject",
        lambda cls: (
            BadKeyTypes if cls is module.KeyTypes else wallet.session.context.inject(cls)
        ),
    )
    monkeypatch.setattr(module, "ERR_UNKNOWN_KEY_TYPE", "Unknown key type %s")
    with pytest.raises(module.WalletError):
        await wallet.assign_kid_to_key(info.verkey, "kid-bad")


@pytest.mark.asyncio
async def test_get_public_did_populates_and_store_did_paths(wallet_env):
    module, wallet, profile = wallet_env

    did_info = await wallet.create_local_did(
        module.SOV, module.ED25519, metadata={"public": True}
    )
    pub = await wallet.get_public_did()
    assert pub and pub.did == did_info.did

    with pytest.raises(module.WalletDuplicateError):
        await wallet.store_did(did_info)

    new_info = module.DIDInfo(
        did="did:sov:newdid",
        verkey=did_info.verkey,
        metadata={},
        method=module.SOV,
        key_type=module.ED25519,
    )
    stored = await wallet.store_did(new_info)
    assert stored.did == "did:sov:newdid"


@pytest.mark.asyncio
async def test_set_public_did_with_didinfo_and_pack_message_errors(wallet_env):
    module, wallet, profile = wallet_env

    did_info = await wallet.create_local_did(module.SOV, module.ED25519)
    info = await wallet.set_public_did(did_info)
    assert info.did == did_info.did

    with pytest.raises(module.WalletError):
        await wallet.pack_message(None, ["v1"])  # message None
    with pytest.raises(module.WalletNotFoundError):
        await wallet.pack_message("{}", ["v1"], from_verkey="missing")


@pytest.mark.asyncio
async def test_get_local_did_and_replace_metadata_errors(wallet_env):
    module, wallet, profile = wallet_env

    with pytest.raises(module.WalletNotFoundError):
        await wallet.get_local_did("")

    with pytest.raises(module.WalletNotFoundError):
        await wallet.replace_local_did_metadata("did:missing", {})


@pytest.mark.asyncio
async def test_create_key_askar_error_non_duplicate(wallet_env, monkeypatch):
    module, wallet, profile = wallet_env

    class LocalErr(Exception):
        def __init__(self, code):
            self.code = code

    monkeypatch.setattr(module, "AskarError", LocalErr)
    monkeypatch.setattr(
        module, "AskarErrorCode", types.SimpleNamespace(DUPLICATE="DUP", INPUT="INPUT")
    )

    async def _raise(*a, **kw):
        raise LocalErr("BUSY")

    monkeypatch.setattr(profile.askar_handle, "insert_key", _raise)

    with pytest.raises(module.WalletError):
        await wallet.create_key(module.ED25519)


@pytest.mark.asyncio
async def test_create_local_did_duplicate_updates_metadata(wallet_env, monkeypatch):
    module, wallet, profile = wallet_env

    did = "did:sov:dupmeta"
    verkey = "samevk"
    async with profile.store.session() as s:
        await s.insert(
            "did",
            did,
            value_json={
                "did": did,
                "method": "sov",
                "verkey": verkey,
                "verkey_type": "ed25519",
                "metadata": {"a": 1},
            },
            tags={"method": "sov", "verkey": verkey, "verkey_type": "ed25519"},
        )

    class LocalErr(Exception):
        def __init__(self, code):
            self.code = code

    monkeypatch.setattr(module, "AskarError", LocalErr)
    monkeypatch.setattr(module, "AskarErrorCode", types.SimpleNamespace(DUPLICATE="DUP"))

    async def _dup(*a, **kw):
        raise LocalErr("DUP")

    monkeypatch.setattr(profile.askar_handle, "insert_key", _dup)

    class _DPV:
        def __init__(self, *_):
            pass

        def validate_key_type(self, *_):
            return None

        def validate_or_derive_did(self, *_):
            return did

    monkeypatch.setattr(module, "DIDParametersValidation", _DPV)

    async def _fetch_key(name, *a, **kw):
        return types.SimpleNamespace(
            key=types.SimpleNamespace(
                get_public_bytes=lambda: b"irrelevant",
                algorithm=types.SimpleNamespace(value="ed25519"),
            )
        )

    monkeypatch.setattr(profile.askar_handle, "fetch_key", _fetch_key)

    monkeypatch.setattr(module, "bytes_to_b58", lambda *_: "samevk")
    updated = await wallet.create_local_did(module.SOV, module.ED25519, metadata={})
    assert updated.did == did
    info = await wallet.get_local_did(did)
    assert info.metadata == {}


@pytest.mark.asyncio
async def test_get_local_did_db_error_mapping(wallet_env, monkeypatch):
    module, wallet, profile = wallet_env

    class BadSess:
        async def __aenter__(self):
            return self

        async def __aexit__(self, et, ev, tb):
            return False

        async def fetch(self, *a, **kw):
            from acapy_agent.database_manager.dbstore import (
                DBStoreError,
                DBStoreErrorCode,
            )

            raise DBStoreError(DBStoreErrorCode.BUSY, "x")

    def _session_factory():
        return BadSess()

    monkeypatch.setattr(profile.store, "session", _session_factory)
    from acapy_agent.storage import kanon_storage as storage_module

    async def _get_record_mock(*a, **kw):
        raise storage_module.StorageNotFoundError("not found")

    monkeypatch.setattr(storage_module.KanonStorage, "get_record", _get_record_mock)
    with pytest.raises(module.WalletError):
        await wallet.get_local_did("did:any")


@pytest.mark.asyncio
async def test_set_public_did_metadata_update_db_error(wallet_env, monkeypatch):
    module, wallet, profile = wallet_env

    did_info = await wallet.create_local_did(module.SOV, module.ED25519, metadata={})

    class BadSess:
        async def __aenter__(self):
            return self

        async def __aexit__(self, et, ev, tb):
            return False

        async def fetch(self, *a, **kw):
            return None

        async def replace(self, *a, **kw):
            from acapy_agent.database_manager.dbstore import (
                DBStoreError,
                DBStoreErrorCode,
            )

            raise DBStoreError(DBStoreErrorCode.BUSY, "fail")

    async def _get_public():
        return did_info

    monkeypatch.setattr(wallet, "get_public_did", _get_public)
    monkeypatch.setattr(profile.store, "session", lambda: BadSess())
    with pytest.raises(module.WalletError):
        await wallet.set_public_did(did_info.did)


@pytest.mark.asyncio
async def test_verify_message_fallback_path(wallet_env, monkeypatch):
    module, wallet, profile = wallet_env

    monkeypatch.setattr(module, "verify_signed_message", lambda **kw: True)
    ok = await wallet.verify_message(b"m", b"s", "3sj3", module.BLS12381G2)
    assert ok is True


@pytest.mark.asyncio
async def test_get_local_did_for_verkey_multi_peer4_choice(wallet_env):
    module, wallet, profile = wallet_env

    verkey = "vx1"
    async with profile.store.session() as s:
        await s.insert(
            "did",
            "did:peer:4longer",
            value_json={
                "did": "did:peer:4longer",
                "method": "sov",
                "verkey": verkey,
                "verkey_type": "ed25519",
                "metadata": {},
            },
            tags={"method": "sov", "verkey": verkey, "verkey_type": "ed25519"},
        )
        await s.insert(
            "did",
            "did:peer:4x",
            value_json={
                "did": "did:peer:4x",
                "method": "sov",
                "verkey": verkey,
                "verkey_type": "ed25519",
                "metadata": {},
            },
            tags={"method": "sov", "verkey": verkey, "verkey_type": "ed25519"},
        )
    got = await wallet.get_local_did_for_verkey(verkey)
    assert got.did == "did:peer:4x"


@pytest.mark.asyncio
async def test_get_public_did_reads_existing_record(wallet_env):
    module, wallet, profile = wallet_env

    did_info = await wallet.create_local_did(module.SOV, module.ED25519)
    async with profile.store.session() as s:
        await s.insert(
            "config",
            "default_public_did",
            value=json.dumps({"did": did_info.did}),
            tags={},
        )
    pub = await wallet.get_public_did()
    assert pub and pub.did == did_info.did


@pytest.mark.asyncio
async def test_replace_local_did_metadata_success(wallet_env):
    module, wallet, profile = wallet_env

    did_info = await wallet.create_local_did(
        module.SOV, module.ED25519, metadata={"a": 1}
    )
    await wallet.replace_local_did_metadata(did_info.did, {"a": 2})
    got = await wallet.get_local_did(did_info.did)
    assert got.metadata == {"a": 2}


@pytest.mark.asyncio
async def test_sign_message_bls_path(wallet_env, monkeypatch):
    module, wallet, profile = wallet_env

    bls_key = wallet.session.askar_handle._keys
    k = types.SimpleNamespace(
        algorithm=module.KeyAlg.BLS12_381_G2,
        get_public_bytes=lambda: b"blspub",
        get_secret_bytes=lambda: b"blssec",
    )
    await profile.askar_handle.insert_key("blsver", k, metadata=json.dumps({}), tags={})
    monkeypatch.setattr(module, "sign_message", lambda **kw: b"blssig")
    sig = await wallet.sign_message(b"m", "blsver")
    assert sig == b"blssig"


@pytest.mark.asyncio
async def test_set_public_did_replaces_existing_record(wallet_env):
    module, wallet, profile = wallet_env

    did1 = await wallet.create_local_did(
        module.SOV, module.ED25519, did="did:sov:one", metadata={}
    )
    did2 = await wallet.create_local_did(
        module.SOV, module.ED25519, did="did:sov:two", metadata={}
    )

    async with profile.store.session() as s:
        await s.insert(
            "config",
            "default_public_did",
            value=json.dumps({"did": did1.did}),
            tags={},
        )

    res = await wallet.set_public_did(did2)
    assert res.did == did2.did

    async with profile.store.session() as s:
        cfg = await s.fetch("config", "default_public_did")
        assert json.loads(cfg.value)["did"] == did2.did
    got2 = await wallet.get_local_did(did2.did)
    assert got2.metadata.get("posted") is True



import types
from dataclasses import dataclass
from typing import Any, Dict, Optional

import pytest


@dataclass
class FakeKeyAlg:
    value: str


class FakeKey:
    _seq = 0

    def __init__(self, algorithm: FakeKeyAlg, public_bytes: bytes, secret_bytes: bytes):
        self.algorithm = algorithm
        self._public = public_bytes
        self._secret = secret_bytes

    @staticmethod
    def generate(alg: Any):
        FakeKey._seq += 1
        pub = f"pub{FakeKey._seq}".encode()
        return FakeKey(
            FakeKeyAlg(alg.value if hasattr(alg, "value") else str(alg)), pub, b"sec"
        )

    @staticmethod
    def from_secret_bytes(alg: Any, secret: bytes):
        return FakeKey(
            FakeKeyAlg(alg.value if hasattr(alg, "value") else str(alg)), b"pub", secret
        )

    @staticmethod
    def from_seed(alg: Any, seed: Any, method: Any = None):
        return FakeKey(
            FakeKeyAlg(alg.value if hasattr(alg, "value") else str(alg)), b"pub", b"sec"
        )

    @staticmethod
    def from_public_bytes(alg: Any, public: bytes):
        return FakeKey(
            FakeKeyAlg(alg.value if hasattr(alg, "value") else str(alg)), public, b""
        )

    def get_public_bytes(self) -> bytes:
        return self._public

    def get_secret_bytes(self) -> bytes:
        return self._secret

    def sign_message(self, message):  # pragma: no cover - trivial passthrough
        return b"sig" + (message if isinstance(message, bytes) else b"".join(message))


class FakeAskarHandle:
    def __init__(self):
        self._keys: Dict[str, Dict[str, Any]] = {}

    async def insert_key(
        self,
        name: str,
        key: FakeKey,
        metadata: Optional[str] = None,
        tags: Optional[dict] = None,
    ):
        if name in self._keys:

            class _Err(Exception):
                def __init__(self):
                    self.code = "DUPLICATE"

            raise _Err()
        self._keys[name] = {"key": key, "metadata": metadata, "tags": tags or {}}

    def update_key(
        self, name: str, tags: Optional[dict] = None, metadata: Optional[str] = None
    ):
        entry = self._keys.get(name)
        if not entry:

            class _Err(Exception):
                def __init__(self):
                    self.code = "NOT_FOUND"

            raise _Err()
        if tags is not None:
            entry["tags"] = tags
        if metadata is not None:
            entry["metadata"] = metadata

    def fetch_key(self, name: str, for_update: bool = False):
        entry = self._keys.get(name)
        if not entry:
            return None
        return types.SimpleNamespace(
            key=entry["key"], metadata=entry["metadata"], tags=entry["tags"]
        )

    def fetch_all_keys(self, tag_filter: dict, limit: int = 2):
        result = []
        for verkey, entry in self._keys.items():
            if all(entry["tags"].get(k) == v for k, v in (tag_filter or {}).items()):
                result.append(
                    types.SimpleNamespace(
                        key=entry["key"], metadata=entry["metadata"], tags=entry["tags"]
                    )
                )
                if len(result) >= limit:
                    break
        return result


class FakeStoreSession:
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return False

    def fetch(self, category: str, name: str, for_update: bool = False):
        return None


class FakeStore:
    def session(self):
        return FakeStoreSession()


class FakeContext:
    def __init__(self, registry: dict):
        self._registry = registry

    def inject(self, cls):
        return self._registry[cls]


class FakeProfile:
    def __init__(self, askar_handle, context, store):
        self.askar_handle = askar_handle
        self.context = context
        self.store = store
        self.profile = types.SimpleNamespace(name="test")


class FakeSession:
    def __init__(self, profile: FakeProfile):
        self.askar_handle = profile.askar_handle
        self.store = profile
        self.context = profile.context
        self.profile = profile

    def inject(self, cls):
        return self.context.inject(cls)


@pytest.fixture
def patched_wallet(monkeypatch):
    from acapy_agent.wallet import kanon_wallet as module

    class _KeyAlg:
        ED25519 = FakeKeyAlg("ed25519")
        X25519 = FakeKeyAlg("x25519")
        P256 = FakeKeyAlg("p256")
        BLS12_381_G2 = FakeKeyAlg("bls12_381_g2")

    monkeypatch.setattr(module, "Key", FakeKey, raising=True)
    monkeypatch.setattr(module, "KeyAlg", _KeyAlg, raising=True)
    monkeypatch.setattr(module, "validate_seed", lambda s: b"seedbytes")
    monkeypatch.setattr(module, "pack_message", lambda to, fk, m: b"packed")
    monkeypatch.setattr(module, "unpack_message", lambda h, em: (b"{ }", "r", "s"))

    key_types = module.KeyTypes()

    askar = FakeAskarHandle()
    context = FakeContext({module.KeyTypes: key_types})
    profile = FakeProfile(askar, context, FakeStore())
    session = FakeSession(profile)

    wallet = module.KanonWallet(session)
    return module, wallet, askar


@pytest.mark.asyncio
async def test_create_key_and_get_signing_key_success(patched_wallet):
    module, wallet, _ = patched_wallet

    info = await wallet.create_key(module.ED25519, metadata={"k": 1}, kid="kid-1")
    assert info.verkey  # base58 string from bytes
    assert info.metadata == {"k": 1}
    assert info.kid == "kid-1"

    fetched = await wallet.get_signing_key(info.verkey)
    assert fetched.verkey == info.verkey
    assert fetched.key_type.key_type == module.ED25519.key_type


@pytest.mark.asyncio
async def test_assign_kid_to_key_and_get_by_kid(patched_wallet):
    module, wallet, askar = patched_wallet

    created = await wallet.create_key(module.ED25519, metadata={})

    updated = await wallet.assign_kid_to_key(created.verkey, "kid-xyz")
    assert updated.kid == "kid-xyz"

    looked = await wallet.get_key_by_kid("kid-xyz")
    assert looked.kid == "kid-xyz"
    assert looked.verkey == created.verkey


@pytest.mark.asyncio
async def test_get_signing_key_not_found_raises(patched_wallet):
    module, wallet, _ = patched_wallet
    with pytest.raises(module.WalletNotFoundError):
        await wallet.get_signing_key("")


@pytest.mark.asyncio
async def test_create_signing_key_wrapper(patched_wallet):
    module, wallet, _ = patched_wallet
    info = await wallet.create_signing_key(module.ED25519, metadata={"m": 1})
    assert info.metadata == {"m": 1}


@pytest.mark.asyncio
async def test_get_key_by_kid_not_found_and_duplicate(patched_wallet):
    module, wallet, _ = patched_wallet
    with pytest.raises(module.WalletNotFoundError):
        await wallet.get_key_by_kid("nope")

    await wallet.create_key(module.ED25519, metadata={}, kid="dup")
    await wallet.create_key(module.ED25519, metadata={}, kid="dup")
    with pytest.raises(module.WalletDuplicateError):
        await wallet.get_key_by_kid("dup")


@pytest.mark.asyncio
async def test_replace_signing_key_metadata_requires_verkey(patched_wallet):
    module, wallet, _ = patched_wallet
    with pytest.raises(module.WalletNotFoundError):
        await wallet.replace_signing_key_metadata("", {"x": 1})


@pytest.mark.asyncio
async def test_sign_message_and_verify_missing_inputs_and_missing_key(patched_wallet):
    module, wallet, _ = patched_wallet
    with pytest.raises(module.WalletError):
        await wallet.sign_message(b"", "vk")
    with pytest.raises(module.WalletError):
        await wallet.sign_message(b"m", "")
    with pytest.raises(module.WalletNotFoundError):
        await wallet.sign_message(b"m", "unknown")

    k = await wallet.create_key(module.ED25519, metadata={})
    sig = await wallet.sign_message(b"m", k.verkey)
    with pytest.raises(module.WalletError):
        await wallet.verify_message(b"m", sig, "", module.ED25519)
    with pytest.raises(module.WalletError):
        await wallet.verify_message(b"", sig, k.verkey, module.ED25519)
    with pytest.raises(module.WalletError):
        await wallet.verify_message(b"m", b"", k.verkey, module.ED25519)


@pytest.mark.asyncio
async def test_pack_message_missing_from_key_raises(patched_wallet):
    module, wallet, _ = patched_wallet
    with pytest.raises(module.WalletNotFoundError):
        await wallet.pack_message("{}", ["vk"], from_verkey="vk")


@pytest.mark.asyncio
async def test_get_signing_key_unknown_key_type_raises(patched_wallet):
    module, wallet, askar = patched_wallet

    await askar.insert_key("vkU", FakeKey(FakeKeyAlg("unknown"), b"pubU", b"sec"))
    with pytest.raises(module.WalletError):
        await wallet.get_signing_key("vkU")



import os
import tempfile

import pytest

from acapy_agent.database_manager.databases.errors import DatabaseError, DatabaseErrorCode
from acapy_agent.database_manager.databases.sqlite_normalized.config import SqliteConfig


@pytest.mark.asyncio
async def test_sqlite_config_provision_open_remove_generic():
    with tempfile.TemporaryDirectory() as tmp:
        db_path = os.path.join(tmp, "unit_generic.db")
        cfg = SqliteConfig(uri=f"sqlite://{db_path}", schema_config="generic")
        pool, profile, path, rel = cfg.provision(profile="p1", recreate=True)
        assert profile == "p1" or profile == "default_profile"
        assert path == db_path
        assert rel == "release_0"
        pool2, profile2, path2, rel2 = cfg.open(profile=profile)
        assert profile2 == profile
        assert path2 == path
        assert rel2 == rel
        assert cfg.remove() is True
        assert cfg.remove() is False


@pytest.mark.asyncio
async def test_sqlite_config_open_missing_file_raises():
    cfg = SqliteConfig(uri="sqlite:///does/not/exist.db")
    with pytest.raises(DatabaseError) as e:
        cfg.open(profile="p")
    assert e.value.code in {
        DatabaseErrorCode.DATABASE_NOT_FOUND,
        DatabaseErrorCode.QUERY_ERROR,
    }



"""Database types and data structures for database manager."""

import json
from enum import Enum
from typing import Optional, Sequence


class KeyAlg(Enum):
    """Enumeration of supported key algorithms."""

    A128GCM = "a128gcm"
    A256GCM = "a256gcm"
    A128CBC_HS256 = "a128cbchs256"
    A256CBC_HS512 = "a256cbchs512"
    A128KW = "a128kw"
    A256KW = "a256kw"
    BLS12_381_G1 = "bls12381g1"
    BLS12_381_G2 = "bls12381g2"
    BLS12_381_G1G2 = "bls12381g1g2"
    C20P = "c20p"
    XC20P = "xc20p"
    ED25519 = "ed25519"
    X25519 = "x25519"
    K256 = "k256"
    P256 = "p256"

    @classmethod
    def from_key_alg(cls, alg: str) -> Optional["KeyAlg"]:
        """Get KeyAlg instance from the algorithm identifier."""
        for cmp_alg in KeyAlg:
            if cmp_alg.value == alg:
                return cmp_alg
        return None


class SeedMethod(Enum):
    """Enumeration of supported seed methods."""

    BlsKeyGen = "bls_keygen"

    @classmethod
    def from_seed_method(cls, method: str) -> Optional["SeedMethod"]:
        """Get SeedMethod instance from the method identifier."""
        for cmp_mth in SeedMethod:
            if cmp_mth.value == method:
                return cmp_mth
        return None


class EntryOperation(Enum):
    """Enumeration of database entry operations."""

    INSERT = 0
    REPLACE = 1
    REMOVE = 2


class Entry:
    """A single result from a store query."""

    _KEYS = ("name", "category", "value", "tags")

    def __init__(self, category: str, name: str, value: str | bytes, tags: dict):
        """Initialize Entry."""
        self._category = category
        self._name = name
        # Store value as string; decode bytes to UTF-8 if necessary
        self._value = value.decode("utf-8") if isinstance(value, bytes) else value
        self._tags = tags

    @property
    def category(self) -> str:
        """Get the entry category."""
        return self._category

    @property
    def name(self) -> str:
        """Get the entry name."""
        return self._name

    @property
    def value(self) -> str:
        """Get the entry value."""
        return self._value

    @property
    def raw_value(self) -> memoryview:
        """Get the entry value as raw bytes."""
        return memoryview(self._value.encode())

    @property
    def value_json(self) -> dict:
        """Get the entry value parsed as JSON."""
        return json.loads(self._value)

    @property
    def tags(self) -> dict:
        """Get the entry tags."""
        return self._tags

    def keys(self):
        """Get the entry keys."""
        return self._KEYS

    def __getitem__(self, key):
        """Get item by key."""
        if key in self._KEYS:
            return getattr(self, key)
        raise KeyError

    def __repr__(self) -> str:
        """Return string representation of Entry."""
        return (
            f"<Entry(category={repr(self.category)}, name={repr(self.name)}, "
            f"value={self.value}, tags={self.tags})>"
        )


class EntryList:
    """A list of Entry objects."""

    def __init__(self, entries: Sequence[Entry], length: int = None):
        """Initialize EntryList."""
        self._entries = entries
        self._length = length if length is not None else len(entries)

    @property
    def handle(self):
        """Get dummy handle for compatibility."""
        return id(self)  # Dummy handle for compatibility

    def __getitem__(self, index) -> Entry:
        """Get entry by index."""
        if not isinstance(index, int) or index < 0 or index >= self._length:
            raise IndexError()
        return self._entries[index]

    def __iter__(self):
        """Return iterator over entries."""
        return IterEntryList(self)

    def __len__(self) -> int:
        """Get length of entry list."""
        return self._length

    def __repr__(self) -> str:
        """Return string representation of EntryList."""
        return f"<EntryList(handle={self.handle}, length={self._length})>"


class IterEntryList:
    """Iterator for EntryList."""

    def __init__(self, list: EntryList):
        """Initialize IterEntryList."""
        self._entries = list._entries
        self._len = list._length
        self._pos = 0

    def __next__(self):
        """Get next entry."""
        if self._pos < self._len:
            entry = self._entries[self._pos]
            self._pos += 1
            return entry
        raise StopIteration



"""Tests for PostgreSQL normalized database provisioning."""
# poetry run python \
# acapy_agent/database_manager/test/test_db_store_postgresql_normalized_provision.py

import asyncio
import json
import logging
import os

import pytest

from acapy_agent.database_manager.dbstore import DBStore

# Skip all tests in this file if POSTGRES_URL env var is not set
pytestmark = pytest.mark.postgres

# Configure logging
logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    handlers=[logging.StreamHandler()],
)

# Define the PostgreSQL connection string
conn_str = os.environ.get(
    "POSTGRES_URL", "postgres://myuser:mypass@localhost:5432/test_normalize"
)
profile_name = "test_normalize"
config = {
    "min_connections": 4,
    "max_connections": 15,
    "connect_timeout": 30.0,
    "max_idle": 5.0,
    "max_lifetime": 3600.0,
    "max_sessions": 7,
}

# Sample connection JSON data
CONNECTION_JSON_1 = {
    "state": "active",
    "their_did": "did:peer:1zQmdgg9s3MwBEZ49QGn2ohLHbg6osFTepqumgL8RNZ2Mxhf",
    "inbound_connection_id": "123456",
}

CONNECTION_JSON_2 = {
    "state": "inactive",
    "their_did": "did:peer:2AbCdEfGhIjKlMn1234567890",
    "inbound_connection_id": "456789",
}

CONNECTION_JSON_3 = {
    "state": "active",
    "their_did": "did:peer:3BcDeFgHiJkLmNoP456789012",
    "inbound_connection_id": "conn_123",
}


async def test_provision():
    """Test provisioning a PostgreSQL database."""
    print("Provisioning the PostgreSQL database...")
    store = await DBStore.provision(
        uri=conn_str,
        key_method=None,
        pass_key=None,
        profile=profile_name,
        recreate=True,  # Ensure fresh database
        release_number="release_0_1",
        schema_config="normalize",
        config=config,
    )
    print(f"Database provisioned at {conn_str}")
    return store


async def test_insert(store):
    """Test inserting test data into the database (people category)."""
    print("Inserting test data (people)...")
    async with store.transaction() as session:
        await session.insert(
            category="people",
            name="person1",
            value=json.dumps({"name": "Alice"}),
            tags={"attr::person.gender": "F", "attr::person.status": "active"},
            expiry_ms=3600000,
        )
        await session.insert(
            category="people",
            name="person2",
            value=json.dumps({"name": "Bob"}),
            tags={"attr::person.gender": "M", "attr::person.status": "inactive"},
            expiry_ms=3600000,
        )
        await session.insert(
            category="people",
            name="person3",
            value=json.dumps({"name": "Charlie"}),
            tags={"attr::person.gender": "F", "attr::person.status": "active"},
            expiry_ms=3600000,
        )
        count = await session.count(category="people")
        print(f"Inserted 3 people, total count: {count}")
        assert count == 3, "Expected 3 entries"


async def test_scan(store):
    """Test scanning with tag filter and pagination (people category)."""
    print("Testing scan with tag filter (people)...")
    tag_filter = json.dumps({"attr::person.status": "active"})
    scan = store.scan(
        category="people", tag_filter=tag_filter, limit=10, offset=0, profile=profile_name
    )
    entries = [entry async for entry in scan]
    print(f"Found {len(entries)} active people")
    assert len(entries) == 2, "Expected 2 active people"
    for entry in entries:
        try:
            value = json.loads(entry.value)
            print(f" - {entry.name}: {value}")
        except json.JSONDecodeError:
            print(f"Failed to parse JSON for {entry.name}: {entry.value}")
            raise

    print("Testing scan with limit and offset (people)...")
    scan_paginated = store.scan(
        category="people", tag_filter=tag_filter, limit=1, offset=1, profile=profile_name
    )
    paginated_entries = [entry async for entry in scan_paginated]
    print(f"Found {len(paginated_entries)} entries with limit=1, offset=1")
    assert len(paginated_entries) == 1, "Expected 1 entry"


async def test_replace(store):
    """Test replacing entries (people category)."""
    print("Testing replace (people)...")
    async with store.transaction() as session:
        await session.insert(
            category="people",
            name="person4",
            value=json.dumps({"name": "David"}),
            tags={"attr::person.gender": "M", "attr::person.status": "active"},
            expiry_ms=3600000,
        )
        await session.replace(
            category="people",
            name="person1",
            value=json.dumps({"name": "Alice Updated"}),
            tags={"attr::person.gender": "F", "attr::person.status": "inactive"},
        )
        entry = await session.fetch(category="people", name="person1")
        print(f"Updated entry: {entry}")
        assert entry.value == json.dumps({"name": "Alice Updated"}), "Value not updated"

        await session.replace(
            category="people",
            name="person4",
            value=json.dumps({"name": "David Updated"}),
            tags={"attr::person.gender": "M", "attr::person.status": "inactive"},
        )
        updated_entry = await session.fetch(category="people", name="person4")
        print(f"Updated entry: {updated_entry}")
        assert updated_entry.value == json.dumps({"name": "David Updated"}), (
            "Value not updated"
        )


async def test_complex_filter(store):
    """Test scanning with a complex tag filter (people category)."""
    print("Testing complex filter (people)...")
    complex_tag_filter = json.dumps(
        {
            "$or": [
                {
                    "$and": [
                        {"attr::person.gender": {"$like": "F"}},
                        {"attr::person.status": "active"},
                    ]
                },
                {"$not": {"attr::person.status": "active"}},
            ]
        }
    )
    scan = store.scan(
        category="people", tag_filter=complex_tag_filter, profile=profile_name
    )
    entries = [entry async for entry in scan]
    print(f"Found {len(entries)} entries with complex filter")
    assert len(entries) == 4, "Expected 4 entries"
    for entry in entries:
        value = json.loads(entry.value)
        print(f" - {entry.name}: {value}")


async def test_insert_connections(store):
    """Test inserting connection data."""
    print("Inserting connection data...")
    async with store.transaction() as session:
        await session.insert(
            category="connection",
            name="conn_1",
            value=json.dumps(CONNECTION_JSON_1),
            tags={},
        )
        await session.insert(
            category="connection",
            name="conn_2",
            value=json.dumps(CONNECTION_JSON_2),
            tags={},
        )
        await session.insert(
            category="connection",
            name="conn_3",
            value=json.dumps(CONNECTION_JSON_3),
            tags={},
        )
        conn_1 = await session.fetch(category="connection", name="conn_1")
        conn_2 = await session.fetch(category="connection", name="conn_2")
        conn_3 = await session.fetch(category="connection", name="conn_3")
        print(f"Inserted conn_1: {json.loads(conn_1.value)}")
        print(f"Inserted conn_2: {json.loads(conn_2.value)}")
        print(f"Inserted conn_3: {json.loads(conn_3.value)}")
        assert conn_1 is not None, "Failed to insert conn_1"
        assert conn_2 is not None, "Failed to insert conn_2"
        assert conn_3 is not None, "Failed to insert conn_3"
        count = await session.count(category="connection")
        print(f"Inserted 3 connections, total count: {count}")
        assert count == 3, "Expected 3 connections"


async def test_scan_connections(store):
    """Test scanning connections with value filter and pagination."""
    print("Testing scan with value filter (connections)...")
    entries = []
    async with store.session() as session:
        all_entries = await session.fetch_all(category="connection")
        for entry in all_entries:
            value = json.loads(entry.value)
            if value.get("state") == "active":
                entries.append(entry)
        print(f"Found {len(entries)} active connections")
        assert len(entries) == 2, "Expected 2 active connections"
        for entry in entries:
            print(f" - {entry.name}: {json.loads(entry.value)}")

    print("Testing scan with limit and offset (connections)...")
    entries_paginated = entries[1:2]  # Simulate offset=1, limit=1
    print(f"Found {len(entries_paginated)} entries with limit=1, offset=1")
    assert len(entries_paginated) == 1, "Expected 1 entry"


async def test_count_connections(store):
    """Test counting connections with a value filter."""
    print("Testing count with value filter (connections)...")
    async with store.session() as session:
        all_entries = await session.fetch_all(category="connection")
        count = sum(
            1 for entry in all_entries if json.loads(entry.value).get("state") == "active"
        )
        print(f"Counted {count} active connections")
        assert count == 2, "Expected 2 active connections"


async def test_replace_connections(store):
    """Test replacing connection entries."""
    print("Testing replace (connections)...")
    async with store.transaction() as session:
        updated_json = CONNECTION_JSON_1.copy()
        updated_json["state"] = "completed"
        await session.replace(
            category="connection", name="conn_1", value=json.dumps(updated_json), tags={}
        )
        updated_entry = await session.fetch(category="connection", name="conn_1")
        print(f"Updated conn_1: {json.loads(updated_entry.value)}")
        assert json.loads(updated_entry.value)["state"] == "completed", (
            "State not updated"
        )

        await session.insert(
            category="connection",
            name="conn_4",
            value=json.dumps(CONNECTION_JSON_1),
            tags={},
        )
        new_entry = await session.fetch(category="connection", name="conn_4")
        print(f"Inserted conn_4: {json.loads(new_entry.value)}")
        assert new_entry is not None, "Insert failed"

        updated_json_4 = CONNECTION_JSON_1.copy()
        updated_json_4["state"] = "inactive"
        await session.replace(
            category="connection",
            name="conn_4",
            value=json.dumps(updated_json_4),
            tags={},
        )
        updated_conn4 = await session.fetch(category="connection", name="conn_4")
        print(f"Updated conn_4: {json.loads(updated_conn4.value)}")
        assert json.loads(updated_conn4.value)["state"] == "inactive", "State not updated"


async def test_remove_connections(store):
    """Test removing connection entries."""
    print("Testing remove (connections)...")
    async with store.transaction() as session:
        await session.remove(category="connection", name="conn_3")
        removed_entry = await session.fetch(category="connection", name="conn_3")
        assert removed_entry is None, "conn_3 should be removed"


async def test_wql_exist_connections(store):
    """Test WQL $exist query for connections."""
    print("Testing WQL $exist query (connections)...")
    async with store.session() as session:
        all_entries = await session.fetch_all(category="connection")
        entries = [
            entry
            for entry in all_entries
            if "inbound_connection_id" in json.loads(entry.value)
        ]
        print(f"Found {len(entries)} connections with 'inbound_connection_id'")
        assert len(entries) == 3, "Expected 3 connections with 'inbound_connection_id'"
        for entry in entries:
            assert "inbound_connection_id" in json.loads(entry.value), (
                "inbound_connection_id should exist"
            )


async def test_fetch_all(store):
    """Test fetching all entries (people category)."""
    print("Fetching all entries (people)...")
    async with store.session() as session:
        entries = await session.fetch_all(category="people")
        print(f"Found {len(entries)} people")
        assert len(entries) == 4, "Expected 4 entries after replace"


async def test_remove_all_people(store):
    """Test removing all people entries."""
    print("Removing all people entries...")
    async with store.transaction() as session:
        deleted_count = await session.remove_all(category="people")
        print(f"Deleted {deleted_count} people entries")
        assert deleted_count == 4, "Expected to delete 4 entries"


async def test_remove_all_connections(store):
    """Test removing all connection entries."""
    print("Testing remove all connections...")
    async with store.transaction() as session:
        deleted_count = await session.remove_all(category="connection")
        print(f"Deleted {deleted_count} connection entries")
        assert deleted_count >= 3, "Expected to delete at least 3 entries"


async def test_non_encrypted():
    """Test provisioning and using a non-encrypted PostgreSQL database."""
    print("Provisioning non-encrypted PostgreSQL database...")
    non_enc_store = await DBStore.provision(
        uri=conn_str,
        key_method=None,
        pass_key=None,
        profile=profile_name,
        recreate=True,
        release_number="release_0_1",
        schema_config="normalize",
        config=config,
    )
    print(f"Non-encrypted database provisioned at {conn_str}")

    async with non_enc_store.transaction() as session:
        await session.insert(
            category="people",
            name="person4",
            value=json.dumps({"name": "David"}),
            tags={"attr::person.gender": "M", "attr::person.status": "active"},
        )
        await session.insert(
            category="people",
            name="person5",
            value=json.dumps({"name": "Eve"}),
            tags={"attr::person.gender": "F", "attr::person.status": "inactive"},
        )
        count = await session.count(category="people")
        print(f"Inserted {count} people")

    tag_filter = json.dumps({"attr::person.status": "active"})
    scan = non_enc_store.scan(
        category="people", tag_filter=tag_filter, profile=profile_name
    )
    entries = [entry async for entry in scan]
    print(f"Found {len(entries)} active people in non-encrypted db")
    assert len(entries) == 1, "Expected 1 active person"

    await test_insert_connections(non_enc_store)
    await test_scan_connections(non_enc_store)
    await test_count_connections(non_enc_store)
    await test_replace_connections(non_enc_store)
    await test_remove_connections(non_enc_store)
    await test_wql_exist_connections(non_enc_store)

    await non_enc_store.close()


async def cleanup():
    """Clean up the PostgreSQL database."""
    print("Cleaning up PostgreSQL database...")
    await DBStore.remove(conn_str)
    print("Database removed")


async def main():
    """Main test function executing all test scenarios."""
    print("Starting PostgreSQL db_store.py test program...")
    await test_provision()

    print("All tests passed successfully!")


if __name__ == "__main__":
    asyncio.run(main())
